# coding: utf8

import clinica.pipelines.engine as cpe


class T1VolumeParcellation(cpe.Pipeline):
    """T1VolumeParcellation - Computation of mean GM concentration for a set of regions

    Returns:
        A clinica pipeline object containing the T1VolumeParcellation pipeline.
    """
    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        default_atlases = ['AAL2', 'LPBA40', 'Neuromorphometrics', 'AICHA', 'Hammers']

        if 'group_id' not in self.parameters.keys():
            raise KeyError('Missing compulsory group_id key in pipeline parameter.')
        if 'atlases' not in self.parameters.keys():
            self.parameters['atlases'] = default_atlases

        check_group_label(self.parameters['group_id'])

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return ['file_list', 'atlas_list']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        pass

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        from colorama import Fore
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import t1_volume_template_tpm_in_mni
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_groups_in_caps_directory, print_images_to_process

        # Check that group already exists
        if not os.path.exists(os.path.join(self.caps_directory, 'groups', 'group-' + self.parameters['group_id'])):
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                '%sGroup %s does not exist. Did you run t1-volume or t1-volume-create-dartel pipeline?%s' %
                (Fore.RED, self.parameters['group_id'], Fore.RESET)
            )

        try:
            gm_mni = clinica_file_reader(self.subjects,
                                         self.sessions,
                                         self.caps_directory,
                                         t1_volume_template_tpm_in_mni(self.parameters['group_id'], 1, True))
        except ClinicaException as e:
            final_error_str = 'Clinica faced error(s) while trying to read files in your CAPS directory.\n'
            final_error_str += str(e)
            raise ClinicaCAPSError(final_error_str)

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True))
        read_parameters_node.inputs.file_list = gm_mni
        read_parameters_node.inputs.atlas_list = self.parameters['atlases']

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('The pipeline will last a few seconds per image.')

        self.connect([
            (read_parameters_node, self.input_node, [('file_list', 'file_list')]),
            (read_parameters_node, self.input_node, [('atlas_list', 'atlas_list')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        from ..t1_volume_parcellation import t1_volume_parcellation_utils as parcellation_utils

        atlas_stats_node = npe.MapNode(nutil.Function(input_names=['in_image',
                                                                   'atlas_list'],
                                                      output_names=['atlas_statistics'],
                                                      function=parcellation_utils.atlas_statistics),
                                       name='atlas_stats_node',
                                       iterfield=['in_image'])
        outputnode = npe.Node(nutil.IdentityInterface(fields=['atlas_statistics']),
                              name='outputnode',
                              mandatory_inputs=True)

        datasink = npe.Node(nio.DataSink(),
                            name='datasink')

        datasink.inputs.base_directory = self.caps_directory
        datasink.inputs.parameterization = True
        datasink.inputs.regexp_substitutions = [
            (r'(.*)(atlas_statistics)/.*/(sub-(.*)_ses-(.*)_T1.*)$',
             r'\1/subjects/sub-\4/ses-\5/t1/spm/dartel/group-' + self.parameters['group_id'] + r'/\2/\3')]

        # Connection
        # ==========
        self.connect([
            (self.input_node,      atlas_stats_node,    [('file_list',    'in_image')]),
            (self.input_node,      atlas_stats_node,    [('atlas_list',    'atlas_list')]),
            (atlas_stats_node,     outputnode,          [('atlas_statistics',  'atlas_statistics')]),
            (outputnode,           datasink,            [('atlas_statistics', 'atlas_statistics')])
        ])



# coding: utf8

import clinica.engine as ce


class T1FreeSurferCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-freesurfer'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Cross-sectional pre-processing of T1w images with FreeSurfer:\n'
                             'http://clinica.run/doc/Pipelines/T1_FreeSurfer/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')

        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-raa", "--recon_all_args",
                              metavar='flag(s)', type=str, default="-qcache",
                              help='Additional flags for recon-all command line '
                                   '(default: --recon_all_args="%(default)s"). '
                                   'Please note that = is compulsory after --recon_all_args/-raa flag '
                                   '(this is not the case for other flags).')

        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments(add_overwrite_flag=True)

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .t1_freesurfer_pipeline import T1FreeSurfer
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'recon_all_args': args.recon_all_args,
        }
        pipeline = T1FreeSurfer(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name,
            overwrite_caps=args.overwrite_outputs
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8


import os
from os import path
import numpy as np

from clinica.pipelines.machine_learning import base, input, algorithm, validation

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class VoxelBasedKFoldDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_folds=10,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):

        super(VoxelBasedKFoldDualSVM, self).__init__(input.CAPSVoxelBasedInput,
                                                     validation.KFoldCV,
                                                     algorithm.DualSVMAlgorithm,
                                                     locals(),
                                                     output_dir)


class VoxelBasedRepKFoldDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 n_folds=10, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):

        super(VoxelBasedRepKFoldDualSVM, self).__init__(input.CAPSVoxelBasedInput,
                                                        validation.RepeatedKFoldCV,
                                                        algorithm.DualSVMAlgorithm,
                                                        locals(),
                                                        output_dir)


class VoxelBasedRepHoldOutDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 test_size=0.3, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17),
                 splits_indices=None):

        super().__init__(input.CAPSVoxelBasedInput,
                         validation.RepeatedHoldOut,
                         algorithm.DualSVMAlgorithm,
                         locals(),
                         output_dir)


class VertexBasedRepHoldOutDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, output_dir, image_type='fdg', fwhm=20,
                 precomputed_kernel=None, n_threads=15, n_iterations=100, test_size=0.3, grid_search_folds=10,
                 balanced=True, c_range=np.logspace(-10, 2, 1000), splits_indices=None):

        super(VertexBasedRepHoldOutDualSVM, self).__init__(input.CAPSVertexBasedInput,
                                                           validation.RepeatedHoldOut,
                                                           algorithm.DualSVMAlgorithm,
                                                           locals(),
                                                           output_dir)


class RegionBasedRepHoldOutDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):

        super(RegionBasedRepHoldOutDualSVM, self).__init__(input.CAPSRegionBasedInput,
                                                           validation.RepeatedHoldOut,
                                                           algorithm.DualSVMAlgorithm,
                                                           locals(),
                                                           output_dir)


class RegionBasedRepHoldOutLogisticRegression(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):

        super(RegionBasedRepHoldOutLogisticRegression, self).__init__(input.CAPSRegionBasedInput,
                                                                      validation.RepeatedHoldOut,
                                                                      algorithm.LogisticReg,
                                                                      locals(),
                                                                      output_dir)


class RegionBasedRepHoldOutRandomForest(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, n_estimators_range=(100, 200, 400),
                 max_depth_range=[None], min_samples_split_range=[2],
                 max_features_range=('auto', 0.25, 0.5), splits_indices=None):

        super(RegionBasedRepHoldOutRandomForest, self).__init__(input.CAPSRegionBasedInput,
                                                                validation.RepeatedHoldOut,
                                                                algorithm.RandomForest,
                                                                locals(),
                                                                output_dir)


class RegionBasedLearningCurveRepHoldOutDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas,
                 output_dir, pvc=None, precomputed_kernel=None, n_threads=15, n_iterations=100, test_size=0.3,
                 n_learning_points=10, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17)):

        super(RegionBasedLearningCurveRepHoldOutDualSVM, self).__init__(input.CAPSRegionBasedInput,
                                                                        validation.LearningCurveRepeatedHoldOut,
                                                                        algorithm.DualSVMAlgorithm,
                                                                        locals(),
                                                                        output_dir)


class VoxelBasedLearningCurveRepHoldOutDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 test_size=0.3, n_learning_points=10, grid_search_folds=10, balanced=True,
                 c_range=np.logspace(-6, 2, 17)):

        super(VoxelBasedLearningCurveRepHoldOutDualSVM, self).__init__(input.CAPSVoxelBasedInput,
                                                                       validation.LearningCurveRepeatedHoldOut,
                                                                       algorithm.DualSVMAlgorithm,
                                                                       locals(),
                                                                       output_dir)


class RegionBasedRepKFoldDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3, n_folds=10,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):

        super(RegionBasedRepKFoldDualSVM, self).__init__(input.CAPSRegionBasedInput,
                                                         validation.RepeatedKFoldCV,
                                                         algorithm.DualSVMAlgorithm,
                                                         locals(),
                                                         output_dir)


class CAPSTsvRepHoldOutDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type,  atlas, dataset,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17), splits_indices=None):

        super(CAPSTsvRepHoldOutDualSVM, self).__init__(input.CAPSTSVBasedInput,
                                                       validation.RepeatedHoldOut,
                                                       algorithm.DualSVMAlgorithm,
                                                       locals(),
                                                       output_dir)


class CAPSTsvRepHoldOutRandomForest(base.MLWorkflow):
    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas, dataset,
                 output_dir, pvc=None, n_threads=15, n_iterations=100, test_size=0.3,
                 grid_search_folds=10, balanced=True, n_estimators_range=(100, 200, 400),
                 max_depth_range=[None], min_samples_split_range=[2],
                 max_features_range=('auto', 0.25, 0.5), splits_indices=None):

        super(CAPSTsvRepHoldOutRandomForest, self).__init__(input.CAPSTSVBasedInput,
                                                            validation.RepeatedHoldOut,
                                                            algorithm.RandomForest,
                                                            locals(),
                                                            output_dir)

# SVM reg


class VoxelBasedREGRepKFoldDualSVM(base.MLWorkflow):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, output_dir, fwhm=0,
                 modulated="on", pvc=None, precomputed_kernel=None, mask_zeros=True, n_threads=15, n_iterations=100,
                 n_folds=10,
                 test_size=0.1, grid_search_folds=10, balanced=True, c_range=np.logspace(-6, 2, 17),
                 splits_indices=None):

        super(VoxelBasedREGRepKFoldDualSVM, self).__init__(input.CAPSTSVBasedInput,
                                                           validation.RepeatedKFoldCV,
                                                           algorithm.DualSVMAlgorithm,
                                                           locals(),
                                                           output_dir)


# TSV

class TsvRepHoldOutRandomForest(base.MLWorkflow):

    def __init__(self, data_tsv, columns, output_dir, n_threads=20, n_iterations=250, test_size=0.2,
                 grid_search_folds=10, balanced=True, n_estimators_range=(100, 200, 400), max_depth_range=[None],
                 min_samples_split_range=[2], max_features_range=('auto', 0.25, 0.5), splits_indices=None,
                 inner_cv=False):

        super(TsvRepHoldOutRandomForest, self).__init__(input.TsvInput,
                                                        validation.RepeatedHoldOut,
                                                        algorithm.RandomForest,
                                                        locals(),
                                                        output_dir)



# coding: utf8


def get_t1_freesurfer_custom_file():
    import os
    custom_file = os.path.join(
        '@subject',
        '@session',
        't1',
        'freesurfer_cross_sectional',
        '@subject_@session',
        'surf',
        '@hemi.thickness.fwhm@fwhm.fsaverage.mgh'
    )
    return custom_file


def get_fdg_pet_surface_custom_file():
    import os
    custom_file = os.path.join(
        '@subject',
        '@session',
        'pet',
        'surface',
        '@subject_@session_task-rest_acq-fdg_pet_space-fsaverage_suvr-pons_pvc-iy_hemi-@hemi_fwhm-@fwhm_projection.mgh'
    )
    return custom_file


def init_input_node(parameters, base_dir, subjects_visits_tsv):
    """Initialize the pipeline.

    This function will:
        - Create `surfstat_results_dir` in `base_dir`/<group_id> for SurfStat;
        - Save pipeline parameters in JSON file;
        - Copy TSV file with covariates;
        - Print begin execution message.
    """
    import os
    import errno
    import json
    import shutil
    from clinica.utils.ux import print_begin_image
    from clinica.pipelines.statistics_surface.statistics_surface_utils import create_glm_info_dictionary

    group_id = 'group-' + parameters['group_label']

    # Create surfstat_results_dir for SurfStat
    surfstat_results_dir = os.path.join(base_dir, group_id)
    try:
        os.makedirs(surfstat_results_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:  # EEXIST: folder already exists
            raise e

    # Save pipeline parameters in JSON file
    glm_dict = create_glm_info_dictionary(subjects_visits_tsv, parameters)
    json_filename = os.path.join(surfstat_results_dir, group_id + '_glm.json')
    with open(json_filename, 'w') as json_file:
        json.dump(glm_dict, json_file, indent=4)

    # Copy TSV file with covariates
    tsv_filename = os.path.join(surfstat_results_dir, group_id + '_covariates.tsv')
    shutil.copyfile(subjects_visits_tsv, tsv_filename)

    # Print begin message
    list_keys = ['AnalysisType', 'Covariates', 'Contrast', 'FWHM', 'ClusterThreshold']
    list_values = [
        parameters['glm_type'],
        parameters['covariates'],
        parameters['contrast'],
        str(parameters['full_width_at_half_maximum']),
        str(parameters['cluster_threshold'])
    ]
    group_id = 'group-' + parameters['group_label']
    print_begin_image(group_id, list_keys, list_values)

    return parameters['group_label'], surfstat_results_dir


def get_string_format_from_tsv(tsv_file):
    """
    Determine string format from TSV file.

    If the TSV file is like:

    participant_id  session_id  sex     group   age
    sub-CLNC0001    ses-M00     Female  CN      71.1
    sub-CLNC0002    ses-M00     Male    CN      81.3
    sub-CLNC0003    ses-M00     Male    CN      75.4

    The columns of the TSV file contains consecutively strings, strings,
    strings, strings and float. The string_format is therefore "%s %s %s %s %f".

    Args:
        tsv_file: TSV file.

    Returns:
        String formatting of the TSV file (e.g. "%s %s %s %s %f")
    """
    import pandas as pd

    demographics_df = pd.read_csv(tsv_file, sep='\t')

    def dtype_to_str_format(dtype):
        """Convert pandas dtypes (e.g. int64) to string format (e.g. %d)"""
        import numpy as np
        if dtype == np.int64:
            str_format = "%d"
        elif dtype == np.float64:
            str_format = "%f"
        elif dtype == np.object:
            str_format = "%s"
        else:
            raise ValueError("Unknown dtype (given: %s)" % dtype)
        return str_format

    list_str_format = []
    for column in demographics_df.columns:
        list_str_format.append(dtype_to_str_format(demographics_df[column].dtype))

    return " ".join(list_str_format)


def covariates_to_design_matrix(contrast, covariates=None):
    """
    Generate design matrix for SurfStat based on the contrast and the optional list of covariates.

    Design matrix "1 + <contrast> + <covariate_1> + ... + <covariate_n>"

    Example:
        >>> from clinica.pipelines.statistics_surface.statistics_surface_utils import covariates_to_design_matrix
        >>> covariates_to_design_matrix('group', 'age sex group')
        1 + group + age + sex
        >>> covariates_to_design_matrix('group', 'age')
        1 + group + age
        >>> covariates_to_design_matrix('group', None)
        1 + group
    """
    if covariates:
        # Convert string to list while handling case where several spaces are present
        list_covariates = list(set(covariates.split(' ')))
        try:
            list_covariates.remove('')
        except ValueError:
            pass

        if contrast in list_covariates:
            design_matrix = "1 + " + " + ".join(covariate for covariate in list_covariates)
        else:
            design_matrix = "1 + " + contrast + " + " + " + ".join(covariate for covariate in list_covariates)
    else:
        design_matrix = "1 + " + contrast

    return design_matrix


def run_matlab(caps_dir,
               output_dir,
               subjects_visits_tsv,
               pipeline_parameters):
    """
    Wrap the call of SurfStat using clinicasurfstat.m Matlab script.

    Args:
        caps_dir (str): CAPS directory containing surface-based features
        output_dir (str): Output directory that will contain outputs of clinicasurfstat.m
        subjects_visits_tsv (str): TSV file containing the GLM information
        pipeline_parameters (dict): parameters of StatisticsSurface pipeline
    """
    import os
    import sys
    from nipype.interfaces.matlab import MatlabCommand, get_matlab_command
    import clinica.pipelines as clinica_pipelines
    from clinica.utils.check_dependency import check_environment_variable
    from clinica.pipelines.statistics_surface.statistics_surface_utils import covariates_to_design_matrix, get_string_format_from_tsv

    path_to_matlab_script = os.path.join(os.path.dirname(clinica_pipelines.__path__[0]), 'lib', 'clinicasurfstat')
    freesurfer_home = check_environment_variable('FREESURFER_HOME', 'FreeSurfer')

    MatlabCommand.set_default_matlab_cmd(
        get_matlab_command()
    )
    matlab = MatlabCommand()
    matlab.inputs.paths = path_to_matlab_script
    matlab.inputs.script = """
    clinicasurfstat('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', %d, '%s', %.3f, '%s', %.3f, '%s', %.3f);
    """ % (
        os.path.join(caps_dir, 'subjects'),
        output_dir,
        subjects_visits_tsv,
        covariates_to_design_matrix(
            pipeline_parameters['contrast'],
            pipeline_parameters['covariates']
        ),
        pipeline_parameters['contrast'],
        get_string_format_from_tsv(subjects_visits_tsv),
        pipeline_parameters['glm_type'],
        pipeline_parameters['group_label'],
        freesurfer_home,
        pipeline_parameters['custom_file'],
        pipeline_parameters['measure_label'],
        'sizeoffwhm', pipeline_parameters['full_width_at_half_maximum'],
        'thresholduncorrectedpvalue', 0.001,
        'thresholdcorrectedpvalue', 0.05,
        'clusterthreshold', pipeline_parameters['cluster_threshold']
        )
    # This will create a file: pyscript.m , the pyscript.m is the default name
    matlab.inputs.mfile = True
    # This will stop running with single thread
    matlab.inputs.single_comp_thread = False
    matlab.inputs.logfile = 'group-' + pipeline_parameters['group_label'] + '_matlab.log'

    # cprint("Matlab logfile is located at the following path: %s" % matlab.inputs.logfile)
    # cprint("Matlab script command = %s" % matlab.inputs.script)
    # cprint("MatlabCommand inputs flag: single_comp_thread = %s" % matlab.inputs.single_comp_thread)
    # cprint("MatlabCommand choose which matlab to use(matlab_cmd): %s" % get_matlab_command())
    matlab.run()

    return output_dir


def create_glm_info_dictionary(tsv_file, pipeline_parameters):
    """Create dictionary containing the GLM information that will be stored in a JSON file."""
    out_dict = {
        'AnalysisType': pipeline_parameters['glm_type'],
        'DesignMatrix':  covariates_to_design_matrix(
            pipeline_parameters['contrast'],
            pipeline_parameters['covariates']
        ),
        'StringFormatTSV': get_string_format_from_tsv(tsv_file),
        'Contrast': pipeline_parameters['contrast'],
        'Covariates': pipeline_parameters['covariates'],
        'GroupLabel': pipeline_parameters['group_label'],
        'FWHM': pipeline_parameters['full_width_at_half_maximum'],
        'ThresholdUncorrectedPvalue': 0.001,
        'ThresholdCorrectedPvalue': 0.05,
        'ClusterThreshold': pipeline_parameters['cluster_threshold']
    }
    return out_dict


def save_to_caps(source_dir, caps_dir, overwrite_caps, pipeline_parameters):
    """Save `source_dir`/ to CAPS folder.

    This function copies outputs of `source_dir`/ to
    `caps_dir`/groups/<group_id>/<statistics>/surfstat_<glm_type>/

    The `source_dir`/ folder should contain the following elements:
        - group-<group_label>_<group_1_or_2>-lt-<group_1_or_2>_measure-<measure>_fwhm-<label>_suffix.ext
    or
        - group-<group_label>_correlation-<label>_contrast-{-|+}_measure-<measure>_fwhm-<label>_suffix.ext
    and
        - group-<group_label>_covariates.tsv
        - group-<group_label>_glm.json

    Raise:
        NotImplementedError: If overwrite_caps=True.
    """
    import os
    import shutil
    from clinica.utils.ux import print_end_image

    group_id = 'group-' + pipeline_parameters['group_label']

    if pipeline_parameters['glm_type'] == "group_comparison":
        surfstat_folder = 'surfstat_' + pipeline_parameters['glm_type']
    elif pipeline_parameters['glm_type'] == "correlation":
        surfstat_folder = 'surfstat_' + pipeline_parameters['glm_type'] + '_analysis'
    else:
        raise NotImplementedError("The other GLM situations have not been implemented in this pipeline.")

    destination_dir = os.path.join(
        os.path.expanduser(caps_dir),
        'groups',
        group_id,
        'statistics',
        surfstat_folder
    )

    if overwrite_caps:
        raise NotImplementedError('save_to_caps(overwrite_caps=True) not implemented')
    shutil.copytree(source_dir, destination_dir, symlinks=True)
    print_end_image(group_id)



# coding: utf8

"""
This module handles Clinica exceptions.
"""


class ClinicaException(Exception):
    """ Base class for Clinica exceptions """


class ClinicaMissingDependencyError(ClinicaException):
    """ Base class for Clinica dependencies errors """


class ClinicaBIDSError(ClinicaException):
    """ Base class for BIDS errors """


class ClinicaCAPSError(ClinicaException):
    """ Base class for CAPS errors """



# coding: utf8


import numpy as np
import pandas as pd
import nibabel as nib
from os.path import join

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez", "Simona Bottani"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def get_caps_t1_list(input_directory, subjects_visits_tsv, group_id, atlas_id):
    """
    path to arrive to the list of the file with the statistics on atlas_id
    Args:
        input_directory:
        subjects_visits_tsv:
        group_id:
        atlas_id:

    Returns:

    """

    from os.path import join
    import pandas as pd

    subjects_visits = pd.io.parsers.read_csv(subjects_visits_tsv, sep='\t')
    if list(subjects_visits.columns.values) != ['participant_id', 'session_id']:
        raise Exception('Subjects and visits file is not in the correct format.')
    subjects = list(subjects_visits.participant_id)
    sessions = list(subjects_visits.session_id)
    image_list = [join(input_directory + '/subjects/' + subjects[i] + '/'
                       + sessions[i] + '/t1/spm/dartel/group-' + group_id + '/atlas_statistics/' + subjects[i] + '_'
                       + sessions[i]+'_T1w_space-'+atlas_id+'_map-graymatter_statistics.tsv')
                  for i in range(len(subjects))]
    return image_list


def get_caps_pet_list(input_directory, subjects_visits_tsv, group_id, atlas_id):
    """

    Args:
        input_directory:
        subjects_visits_tsv:
        group_id:
        atlas_id:

    Returns:

    """

    subjects_visits = pd.io.parsers.read_csv(subjects_visits_tsv, sep='\t')
    if list(subjects_visits.columns.values) != ['participant_id', 'session_id']:
        raise Exception('Subjects and visits file is not in the correct format.')
    subjects = list(subjects_visits.participant_id)
    sessions = list(subjects_visits.session_id)
    image_list = [join(input_directory, 'analysis-series-default/subjects/' + subjects[i] + '/'
                       + sessions[i] + '/pet/atlas_statistics/' + subjects[i] + '_' + sessions[i]
                       + '_space-' + atlas_id + '_map-fdgstatistic2.tsv')
                  for i in range(len(subjects))]
    return image_list


def load_data(image_list, subjects):
    """

    Args:
        image_list:
        subjects:

    Returns:

    """

    subj_average = []
    all_vector = np.array([])
    read_file = pd.io.parsers.read_csv(image_list[0], sep='\t', usecols=[2], header=0)
    read_file = read_file.mean_scalar
    data = np.zeros((len(subjects), len(read_file)))

    for i in range(len(image_list)):
        tsv_file = pd.io.parsers.read_csv(image_list[i], sep='\t', usecols=[2], header=0)
        subj_average = tsv_file.mean_scalar
        all_vector = np.append(all_vector, subj_average)
    data_temp = np.split(all_vector, len(image_list))

    for i in range(len(image_list)):
        for j in range(len(subj_average)):
            data[i][j] = data_temp[i][j]
    return data


def features_weights(image_list, dual_coefficients, sv_indices, scaler=None):
    """

    Args:
        image_list:
        dual_coefficients:
        sv_indices:
        scaler:

    Returns:

    """

    if len(sv_indices) != len(dual_coefficients):
        print("Length dual coefficients: " + str(len(dual_coefficients)))
        print("Length indices: " + str(len(sv_indices)))
        raise ValueError('The number of support vectors indices and the number of coefficients must be the same.')

    if len(image_list) == 0:
        raise ValueError('The number of images must be greater than 0.')

    sv_images = [image_list[i] for i in sv_indices]

    shape = pd.io.parsers.read_csv(sv_images[0], sep='\t', usecols=[2], header=0)
    weights = np.zeros(len(shape))

    for i in range(len(sv_images)):
        subj = pd.io.parsers.read_csv(sv_images[i], sep='\t', usecols=[2], header=0)
        subj_data = subj.mean_scalar
        weights += dual_coefficients[i] * subj_data

    return weights


def weights_to_nifti(weights, atlas, output_filename):
    """

    Args:
        atlas:
        weights:
        output_filename:

    Returns:

    """

    from os.path import join, split, realpath

    from clinica.utils.atlas import AtlasAbstract

    atlas_path = None
    atlas_classes = AtlasAbstract.__subclasses__()
    for atlas_class in atlas_classes:
        if atlas_class.get_name_atlas() == atlas:
            atlas_path = atlas_class.get_atlas_labels()

    if not atlas_path:
        raise ValueError('Atlas path not found for atlas name ' + atlas)

    atlas_image = nib.load(atlas_path)
    atlas_data = atlas_image.get_data()
    labels = list(set(atlas_data.ravel()))
    output_image_weights = np.array(atlas_data, dtype='f')
    for i, n in enumerate(labels):
        index = np.array(np.where(atlas_data == n))
        output_image_weights[index[0, :], index[1, :], index[2, :]] = weights[i]

    affine = atlas_image.get_affine()
    header = atlas_image.get_header()
    output_image = nib.Nifti1Image(output_image_weights, affine, header)
    nib.save(output_image, output_filename)



# coding: utf8


import os
import contextlib


def csv_to_tsv(file, path_out):
    import os

    fich = open(file, 'r')
    out = open(os.path.join(path_out, 'ida.tsv'), 'w')
    s = ''
    for line in fich.readlines():
        line_mod = ''
        line = line.split(',')
        for elt in line:
            line_mod += elt + '\t'
        line_mod = line_mod[:-1]
        s += line_mod

    out.write(s)
    out.close()
    fich.close()


def first_process(path_out):
    import os

    path_ida = os.path.join(path_out, 'ida.tsv')

    def get_transf_date(pat_name):
        dates = []
        for elt in pat_dic[pat]:
            line = elt.split('\t')
            date = line[5]
            if date not in dates:
                dates.append(date)
            val = float('inf')
            for date in dates:
                month = int(date.split('/')[0]) + 12 * int(date.split('/')[2])
                if month < val:
                    low_date = date
                    val = month

            transf_date = dict()
            for date in dates:
                transf_date[date] = int(date.split('/')[0]) + 12 * int(date.split('/')[2]) - val
        return transf_date

    def modif_line(pat_name, final_sol):
        transf_date = get_transf_date(pat_name)
        for line in pat_dic[pat]:
            line = line.split('\t')
            line[4] = 'Month ' + str(transf_date[line[5]])

            for elt in line:
                final_sol += elt + '\t'
            final_sol = final_sol[:-1]
        return final_sol

    fich = open(path_ida, 'r')
    final_sol = fich.readline()
    patients = []
    pat_dic = dict()
    for line in fich.readlines():
        pat = line.split('\t')[0]
        if pat not in patients:
            patients.append(pat)
            pat_dic[pat] = [str(line)]
        else:
            pat_dic[pat].extend([str(line)])

    for pat in patients:
        final_sol = modif_line(pat, final_sol)

    fich.close()
    fich = open(path_ida, 'w')
    fich.write(final_sol)
    fich.close()


def second_process(path_out):
    import pandas as pd
    import os

    path_ida = os.path.join(path_out, 'ida.tsv')

    dfIda = pd.read_csv(path_ida, sep='\t')
    pat_list = list(set(list(dfIda['Subject ID'])))
    pat_list.sort()

    dfFinal = dfIda.iloc[0:0]

    for pat in pat_list:

        dfPat = dfIda[dfIda['Subject ID'] == pat]
        out1 = list(set(list(dfPat['Visit'])))
        out1 = [int(month.split(' ')[-1]) for month in out1]
        out1.sort()

        if pat not in ['1_S_0067', '1_S_0119', '1_S_0091', '1_S_0093', '2_S_0010']:
            map = [i - i % 6 if (i % 6 < 6 - i % 6) else i + 6 - i % 6 for i in out1]
        else:
            map = [i - i % 6 if (i % 6 <= 6 - i % 6) else i + 6 - i % 6 for i in out1]

        if pat == '1_S_0030':
            map = [0, 6, 12]
        if pat == '1_S_0316':
            map = [0, 6, 12]
        if pat == '1_S_0334':
            map = [0, 12, 18, 24]
        if pat == '1_S_0349':
            map = [0, 6]
        if pat == '1_S_0354':
            map = [0, 6]

        # Prints the cases where the mapping process creates an overlap (month 0, month 1 will be both be mapped to month 0 for instance)
        # if len(map) != len(set(map)):
        #     print(out1)
        #     print(map)
        #     print('-------------')

        # Prints the cases where an overlap creates a conflict
        # if pat in ['1_S_0030', '1_S_0316', '1_S_0334', '1_S_0349', '1_S_0354']:
        #     print(out1)
        #     print(map)
        #     print('----------')

        dfPat['Visit'] = dfPat['Visit'].apply(lambda x: 'Month ' + str(map[out1.index(int(x.split(' ')[-1]))]))
        dfFinal = dfFinal.append(dfPat)

    def write(df, path, name):
        import os

        name = os.path.join(path, name) + '.tsv'
        df.to_csv(sep='\t', path_or_buf=name, index=False)

    write(dfFinal, path_out, 'ida')


def supress_stdout(func):
    def wrapper(*a, **ka):
        with open(os.devnull, 'w') as devnull:
            with contextlib.redirect_stdout(devnull):
                func(*a, **ka)

    return wrapper


@supress_stdout
def process_ida(path_ida, path_out=None):
    """
    Builds 'ida.tsv' from 'idaSearch_all.csv'

    Args:
        path_ida: Path to the original ida file 'idaSearch_all.csv'
        path_out: Path to the output directory for 'ida.tsv', path_out is None, the ouput directory will be the same as the orginal ida file
    """
    import os

    if path_out is None:
        path_out = path_ida.split('/')[:-1]
        path_out = '/' + os.path.join(*path_out)

    # Process from original csv to tsv
    csv_to_tsv(path_ida, path_out)

    # First processing on ida.tsv, recomputes the 'Visit' column to avoid overlaps in IRM visits
    first_process(path_out)

    # Second processing on ida.tsv, maps the 'Visit' column to multiples of 6, still avoiding overlaps
    second_process(path_out)



# coding: utf8

import clinica.engine as ce


class OasisToBidsCLI(ce.CmdParser):
    def define_name(self):
        """Define the sub-command name to run this command."""
        self._name = 'oasis-to-bids'

    def define_description(self):
        """Define a description of this command."""
        self._description = 'Convert OASIS (http://oasis-brains.org/) into BIDS.'

    def define_options(self):
        """Define the sub-command arguments."""
        self._args.add_argument("dataset_directory",
                                help='Path to the OASIS images directory.')
        self._args.add_argument("clinical_data_directory",
                                help='Path to the OASIS clinical data directory.')
        self._args.add_argument("bids_directory",
                                help='Path to the BIDS directory.')

    def run_command(self, args):
        """Run the converter with defined args."""
        from clinica.iotools.converters.oasis_to_bids.oasis_to_bids import OasisToBids

        oasis_to_bids = OasisToBids()
        oasis_to_bids.convert_images(args.dataset_directory, args.bids_directory)
        oasis_to_bids.convert_clinical_data(args.clinical_data_directory, args.bids_directory)



# coding: utf8

import clinica.engine as ce


class SpatialSVMCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'machinelearning-prepare-spatial-svm'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Prepare input data for SVM with spatial and anatomical regularization:\n'
                             'http://clinica.run/doc/MachineLeaning_PrepareSpatialSVM')

    def define_options(self):
        """Define the sub-command arguments."""
        from colorama import Fore
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_label",
                                  help='User-defined identifier for the provided group of subjects.')
        clinica_comp.add_argument("orig_input_data",
                                  help='''Origin of input data. Type
                                       't1-volume' to use gray matter maps or
                                       'pet-volume' to use SUVr maps.''',
                                  choices=['t1-volume', 'pet-volume'],
                                  )
        # Optional arguments
        optional_pet = self._args.add_argument_group(
            '%sPipeline options if you use inputs from pet-volume pipeline%s' %
            (Fore.BLUE, Fore.RESET)
        )
        optional_pet.add_argument("-pt", "--pet_tracer",
                                  default='fdg',
                                  help='PET tracer. Can be fdg or av45 (default: --pet_tracer %(default)s)')
        optional_pet.add_argument("-no_pvc", "--no_pvc",
                                  action='store_true', default=False,
                                  help="Force the use of non PVC PET data (by default, PVC PET data are used)")
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-fwhm", "--full_width_half_maximum",
                              type=float, metavar='N', default=4,
                              help='Amount of regularization (in mm). In practice, we found the default value '
                                   '(--full_width_half_maximum %(default)s) to be optimal. We therefore '
                                   'do not recommend to change it unless you have a specific reason to do so.')

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .spatial_svm_pipeline import SpatialSVM
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'group_label': args.group_label,
            'orig_input_data': args.orig_input_data,
            'pet_tracer': args.pet_tracer,
            'no_pvc': args.no_pvc,
            'fwhm': args.fwhm,
        }
        pipeline = SpatialSVM(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8


import clinica.engine as ce
from colorama import Fore

class DeepLearningPrepareDataCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'deeplearning-prepare-data'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Prepare data generated Clinica for PyTorch with Tensor extraction:\n'
                             'http://clinica.run/doc/Pipelines/DeepLearning_PrepareData/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES

        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id...)
        # Most of the time, you will want to read your pipeline inputs into
        # a BIDS and/or CAPS directory. If your pipeline does not require BIDS input,
        # simply remove the two lines involving the BIDS directory.
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("extract_method",
                help='''Format of the extracted features. Three options:
                'image' to convert to PyTorch tensor the complete 3D image,
                'patch' to extract 3D volumetric patches and 
                'slice' to extract 2D slices from the image.
                By default the features are extracted from the cropped image.''',
                choices=['image', 'slice', 'patch'], default='image'
                )

        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument('-uui', '--use_uncropped_image',
                              help='''Use the uncropped image instead of the
                              cropped image generated by t1-linear.''',
                              default=False, action="store_true"
                              )

        optional_patch = self._args.add_argument_group(
            "%sPipeline options if you chose ‘patch’ extraction%s" % (Fore.BLUE, Fore.RESET)
        )
        optional_patch.add_argument(
                '-ps', '--patch_size',
                help='''Patch size (default: --patch_size 50).''',
                type=int, default=50
                )
        optional_patch.add_argument(
                '-ss', '--stride_size',
                help='''Stride size (default: --stride_size 50).''',
                type=int, default=50
                )

        optional_slice = self._args.add_argument_group(
            "%sPipeline options if you chose ‘slice’ extraction%s" % (Fore.BLUE, Fore.RESET)
        )
        optional_slice.add_argument(
                '-sd', '--slice_direction',
                help='''Slice direction. Three options:
                '0' -> Sagittal plane,
                '1' -> Coronal plane or
                '2' -> Axial plane
                (default: sagittal plane i.e. --slice_direction 0)''',
                type=int, default=0
                )
        optional_slice.add_argument(
                '-sm', '--slice_mode',
                help='''Slice mode. Two options: 'rgb' to save the slice in
                three identical channels, ‘single’ to save the slice in a
                single channel (default: --slice_mode rgb).''',
                choices=['rgb', 'single'], default='rgb'
                )

        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .deeplearning_prepare_data_pipeline import DeepLearningPrepareData
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            # Add your own pipeline parameters here to use them inside your
            # pipeline. See the file `deeplearning_prepare_data_pipeline.py` to
            # see an example of use.
            'extract_method': args.extract_method,
            'patch_size': args.patch_size,
            'stride_size': args.stride_size,
            'slice_direction': args.slice_direction,
            'slice_mode': args.slice_mode,
            'use_uncropped_image': args.use_uncropped_image,
        }

        pipeline = DeepLearningPrepareData(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

"""
This module gathers formatted messages that are displayed when running Clinica.

These functions are mainly called by the pipelines.
"""

LINES_TO_DISPLAY = 25


def print_images_to_process(list_participant_id, list_session_id):
    """Print which images will be processed by the pipeline."""
    from .stream import cprint
    from clinica.utils.participant import get_unique_subjects
    unique_participants, sessions_per_participant = get_unique_subjects(list_participant_id, list_session_id)

    cprint('The pipeline will be run on the following %s image(s):' % len(list_participant_id))
    for i in range(0, min(len(unique_participants), LINES_TO_DISPLAY)):
        sessions_i_th_participant = ', '.join(s_id for s_id in sessions_per_participant[i])
        cprint("\t%s | %s," % (unique_participants[i], sessions_i_th_participant))

    if len(unique_participants) > LINES_TO_DISPLAY:
        cprint("\t...")
        sessions_last_participant = ', '.join(s_id for s_id in sessions_per_participant[-1])
        cprint("\t%s | %s" % (unique_participants[-1], sessions_last_participant))


def print_begin_image(image_id, list_keys=None, list_values=None):
    """Print begin run pipeline message for a given image `image_id`."""
    import datetime
    from colorama import Fore
    from .stream import cprint

    if list_keys is not None:
        assert(len(list_keys) == len(list_values))

    begin_message = 'Running pipeline for %s' % (image_id.replace('_', ' | '))
    if list_keys and list_values:
        begin_message += ' ('
        begin_message += ', '.join(key + ' = ' + value for key, value in zip(list_keys, list_values))
        begin_message += ')'
    now = datetime.datetime.now().strftime('%H:%M:%S')
    cprint('%s[%s]%s %s' % (Fore.BLUE, now, Fore.RESET, begin_message))


def print_end_image(image_id):
    """Print end pipeline message for a given image `image_id`."""
    import datetime
    from colorama import Fore
    from .stream import cprint

    end_message = '%s has completed' % image_id.replace('_', ' | ')
    now = datetime.datetime.now().strftime('%H:%M:%S')
    cprint('%s[%s]%s %s' % (Fore.GREEN, now, Fore.RESET, end_message))


def print_end_pipeline(cli_name, working_directory, working_directory_was_specified):
    """Print end pipeline message after its execution."""
    import os
    import datetime
    from colorama import Fore
    from .stream import cprint
    from os.path import abspath

    now = datetime.datetime.now().strftime('%H:%M:%S')
    if working_directory_was_specified:
        cprint('%s[%s]%s The %s pipeline has completed. You can now delete the working directory (%s).' %
               (Fore.GREEN, now, Fore.RESET, cli_name, os.path.join(working_directory, cli_name)))
    else:
        cprint('%s[%s]%s The %s pipeline has completed. Working directory was automatically deleted.' %
               (Fore.GREEN, now, Fore.RESET, cli_name))


def print_failed_images(cli_name, image_ids):
    """Print missing images in CAPS folder after a RuntimeError from Nipype."""
    import datetime
    from colorama import Fore
    from .filemanip import extract_subjects_sessions_from_filename
    from .stream import cprint
    from clinica.utils.participant import get_unique_subjects
    list_participant_id, list_session_id = extract_subjects_sessions_from_filename(image_ids)
    unique_participants, sessions_per_participant = get_unique_subjects(list_participant_id, list_session_id)

    now = datetime.datetime.now().strftime('%H:%M:%S')
    cprint('\n%s[%s] The %s pipeline finished with errors.%s\n' %
           (Fore.RED, now, cli_name, Fore.RESET))
    cprint('%sCAPS outputs were not found for %s image(s):%s' %
           (Fore.RED, len(image_ids), Fore.RESET))
    for i in range(0, min(len(unique_participants), LINES_TO_DISPLAY)):
        sessions_i_th_participant = ', '.join(s_id for s_id in sessions_per_participant[i])
        cprint("\t%s%s | %s%s" % (Fore.RED, unique_participants[i], sessions_i_th_participant, Fore.RESET))

    if len(unique_participants) > LINES_TO_DISPLAY:
        cprint("\t...")
        sessions_last_participant = ', '.join(s_id for s_id in sessions_per_participant[-1])
        cprint("\t%s%s | %s%s" % (Fore.RED, unique_participants[-1], sessions_last_participant, Fore.RESET))


def print_crash_files_and_exit(log_file, working_directory):
    """Print command(s) to type in order to extract details after a Nipype RuntimeError and exit with an exception."""
    from colorama import Fore
    from .filemanip import extract_crash_files_from_log_file
    from .exceptions import ClinicaException
    from .stream import cprint

    cprint('%s\nError details can be found by opening the crash file(s) with the following command(s):%s' %
           (Fore.YELLOW, Fore.RESET))

    crash_files = extract_crash_files_from_log_file(log_file)
    for file in crash_files:
        cprint('%s- nipypecli crash %s%s' %
               (Fore.YELLOW, file,  Fore.RESET))

    cprint('%s\n'
           'If your pipeline crashed due to lack of space of network issues, '
           're-run the pipeline with the working directory (-wd %s).\n'
           'Known issues are displayed here: http://www.clinica.run/doc/InteractingWithClinica/#known-issues\n'
           'Otherwise, you can delete it.%s' %
           (Fore.YELLOW, working_directory, Fore.RESET))
    # Force the display of "Documentation can be found..."
    raise ClinicaException('')


def print_groups_in_caps_directory(caps_directory):
    """Print group IDs based on `caps_directory`/groups folder."""
    from .group import extract_group_ids
    from .stream import cprint

    group_ids = extract_group_ids(caps_directory)
    if group_ids == ['']:
        cprint('No group was found in CAPS directory')
    else:
        cprint("Groups that exist in your CAPS directory are %s." %
               ', '.join(g_id.replace('group-', '') for g_id in group_ids))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os.path as op

from nipype.interfaces.base import traits, TraitedSpec, File
from nipype.interfaces.mrtrix3.base import MRTrix3BaseInputSpec, MRTrix3Base


class TractographyInputSpec(MRTrix3BaseInputSpec):
    sph_trait = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        traits.Float,
        argstr='%f,%f,%f,%f')

    in_file = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=-2,
        desc='input file to be processed')

    out_file = File(
        'tracked.tck',
        argstr='%s',
        mandatory=True,
        position=-1,
        usedefault=True,
        desc='output file containing tracks')

    algorithm = traits.Enum(
        'iFOD2',
        'FACT',
        'iFOD1',
        'Nulldist',
        'SD_Stream',
        'Tensor_Det',
        'Tensor_Prob',
        usedefault=True,
        argstr='-algorithm %s',
        desc='tractography algorithm to be used')

    # ROIs processing options
    roi_incl = traits.Either(
        File(exists=True),
        sph_trait,
        argstr='-include %s',
        desc=('specify an inclusion region of interest, streamlines must'
              ' traverse ALL inclusion regions to be accepted'))
    roi_excl = traits.Either(
        File(exists=True),
        sph_trait,
        argstr='-exclude %s',
        desc=('specify an exclusion region of interest, streamlines that'
              ' enter ANY exclude region will be discarded'))
    roi_mask = traits.Either(
        File(exists=True),
        sph_trait,
        argstr='-mask %s',
        desc=('specify a masking region of interest. If defined,'
              'streamlines exiting the mask will be truncated'))

    # Streamlines tractography options
    step_size = traits.Float(
        argstr='-step %f',
        desc=('set the step size of the algorithm in mm (default is 0.1'
              ' x voxelsize; for iFOD2: 0.5 x voxelsize)'))
    angle = traits.Float(
        argstr='-angle %f',
        desc=('set the maximum angle between successive steps (default '
              'is 90deg x stepsize / voxelsize)'))
    select = traits.Int(
        argstr='-select %d',
        desc=('set the desired number of tracks. The program will continue'
              ' to generate tracks until this number of tracks have been '
              'selected and written to the output file'))
    max_tracks = traits.Int(
        argstr='-maxnum %d',
        desc=('set the maximum number of tracks to generate. The program '
              'will not generate more tracks than this number, even if '
              'the desired number of tracks hasn\'t yet been reached '
              '(default is 100 x number)'))
    max_length = traits.Float(
        argstr='-maxlength %f',
        desc=('set the maximum length of any track in mm (default is '
              '100 x voxelsize)'))
    min_length = traits.Float(
        argstr='-minlength %f',
        desc=('set the minimum length of any track in mm (default is '
              '5 x voxelsize)'))
    cutoff = traits.Float(
        argstr='-cutoff %f',
        desc=('set the FA or FOD amplitude cutoff for terminating '
              'tracks (default is 0.1)'))
    cutoff_init = traits.Float(
        argstr='-initcutoff %f',
        desc=('set the minimum FA or FOD amplitude for initiating '
              'tracks (default is the same as the normal cutoff)'))
    n_trials = traits.Int(
        argstr='-trials %d',
        desc=('set the maximum number of sampling trials at each point'
              ' (only used for probabilistic tracking)'))
    unidirectional = traits.Bool(
        argstr='-unidirectional',
        desc=('track from the seed point in one direction only '
              '(default is to track in both directions)'))
    init_dir = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        argstr='-initdirection %f,%f,%f',
        desc=('specify an initial direction for the tracking (this '
              'should be supplied as a vector of 3 comma-separated values'))
    noprecompt = traits.Bool(
        argstr='-noprecomputed',
        desc=('do NOT pre-compute legendre polynomial values. Warning: this '
              'will slow down the algorithm by a factor of approximately 4'))
    power = traits.Int(
        argstr='-power %d',
        desc=('raise the FOD to the power specified (default is 1/nsamples)'))
    n_samples = traits.Int(
        4, usedefault=True,
        argstr='-samples %d',
        desc=('set the number of FOD samples to take per step for the 2nd '
              'order (iFOD2) method'))
    use_rk4 = traits.Bool(
        argstr='-rk4',
        desc=('use 4th-order Runge-Kutta integration (slower, but eliminates'
              ' curvature overshoot in 1st-order deterministic methods)'))
    stop = traits.Bool(
        argstr='-stop',
        desc=('stop propagating a streamline once it has traversed all '
              'include regions'))
    downsample = traits.Float(
        argstr='-downsample %f',
        desc='downsample the generated streamlines to reduce output file size')

    # Anatomically-Constrained Tractography options
    act_file = File(
        exists=True,
        argstr='-act %s',
        desc=('use the Anatomically-Constrained Tractography framework during'
              ' tracking; provided image must be in the 5TT '
              '(five - tissue - type) format'))
    backtrack = traits.Bool(
        argstr='-backtrack', desc='allow tracks to be truncated')

    crop_at_gmwmi = traits.Bool(
        argstr='-crop_at_gmwmi',
        desc=('crop streamline endpoints more '
              'precisely as they cross the GM-WM interface'))

    # Tractography seeding options
    seed_sphere = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        traits.Float,
        argstr='-seed_sphere %f,%f,%f,%f',
        desc='spherical seed')
    seed_image = File(
        exists=True,
        argstr='-seed_image %s',
        desc='seed streamlines entirely at random within mask')
    seed_rnd_voxel = traits.Tuple(
        File(exists=True),
        traits.Int(),
        argstr='-seed_random_per_voxel %s %d',
        xor=['seed_image', 'seed_grid_voxel'],
        desc=('seed a fixed number of streamlines per voxel in a mask '
              'image; random placement of seeds in each voxel'))
    seed_grid_voxel = traits.Tuple(
        File(exists=True),
        traits.Int(),
        argstr='-seed_grid_per_voxel %s %d',
        xor=['seed_image', 'seed_rnd_voxel'],
        desc=('seed a fixed number of streamlines per voxel in a mask '
              'image; place seeds on a 3D mesh grid (grid_size argument '
              'is per axis; so a grid_size of 3 results in 27 seeds per'
              ' voxel)'))
    seed_rejection = File(
        exists=True,
        argstr='-seed_rejection %s',
        desc=('seed from an image using rejection sampling (higher '
              'values = more probable to seed from'))
    seed_gmwmi = File(
        exists=True,
        argstr='-seed_gmwmi %s',
        requires=['act_file'],
        desc=('seed from the grey matter - white matter interface (only '
              'valid if using ACT framework)'))
    seed_dynamic = File(
        exists=True,
        argstr='-seed_dynamic %s',
        desc=('determine seed points dynamically using the SIFT model '
              '(must not provide any other seeding mechanism). Note that'
              ' while this seeding mechanism improves the distribution of'
              ' reconstructed streamlines density, it should NOT be used '
              'as a substitute for the SIFT method itself.'))
    max_seed_attempts = traits.Int(
        argstr='-max_seed_attempts %d',
        desc=('set the maximum number of times that the tracking '
              'algorithm should attempt to find an appropriate tracking'
              ' direction from a given seed point'))
    out_seeds = File(
        'out_seeds.nii.gz', usedefault=True,
        argstr='-output_seeds %s',
        desc=('output the seed location of all successful streamlines to'
              ' a file'))


class TractographyOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='the output filtered tracks')
    out_seeds = File(
        desc=('output the seed location of all successful'
              ' streamlines to a file'))


class Tractography(MRTrix3Base):
    """
    Performs streamlines tractography after selecting the appropriate
    algorithm.

    .. [FACT] Mori, S.; Crain, B. J.; Chacko, V. P. & van Zijl,
      P. C. M. Three-dimensional tracking of axonal projections in the
      brain by magnetic resonance imaging. Annals of Neurology, 1999,
      45, 265-269

    .. [iFOD1] Tournier, J.-D.; Calamante, F. & Connelly, A. MRtrix:
      Diffusion tractography in crossing fiber regions. Int. J. Imaging
      Syst. Technol., 2012, 22, 53-66

    .. [iFOD2] Tournier, J.-D.; Calamante, F. & Connelly, A. Improved
      probabilistic streamlines tractography by 2nd order integration
      over fibre orientation distributions. Proceedings of the
      International Society for Magnetic Resonance in Medicine, 2010, 1670

    .. [Nulldist] Morris, D. M.; Embleton, K. V. & Parker, G. J.
      Probabilistic fibre tracking: Differentiation of connections from
      chance events. NeuroImage, 2008, 42, 1329-1339

    .. [Tensor_Det] Basser, P. J.; Pajevic, S.; Pierpaoli, C.; Duda, J.
      and Aldroubi, A. In vivo fiber tractography using DT-MRI data.
      Magnetic Resonance in Medicine, 2000, 44, 625-632

    .. [Tensor_Prob] Jones, D. Tractography Gone Wild: Probabilistic Fibre
      Tracking Using the Wild Bootstrap With Diffusion Tensor MRI. IEEE
      Transactions on Medical Imaging, 2008, 27, 1268-1274


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> tk = mrt.Tractography()
    >>> tk.inputs.in_file = 'fods.mif'
    >>> tk.inputs.roi_mask = 'mask.nii.gz'
    >>> tk.inputs.seed_sphere = (80, 100, 70, 10)
    >>> tk.cmdline                               # doctest: +ELLIPSIS
    'tckgen -algorithm iFOD2 -samples 4 -output_seeds out_seeds.nii.gz \
-mask mask.nii.gz -seed_sphere \
80.000000,100.000000,70.000000,10.000000 fods.mif tracked.tck'
    >>> tk.run()                                 # doctest: +SKIP
    """

    _cmd = 'tckgen'
    input_spec = TractographyInputSpec
    output_spec = TractographyOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if 'roi_' in name and isinstance(value, tuple):
            value = ['%f' % v for v in value]
            return trait_spec.argstr % ','.join(value)

        return super(Tractography, self)._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = op.abspath(self.inputs.out_file)
        return outputs



# coding: utf8

import clinica.pipelines.engine as cpe
from nipype import config

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class DwiDti(cpe.Pipeline):
    """DTI-based processing of DWI datasets.

    Returns:
        A clinica pipeline object containing the DwiDti pipeline.

    Raises:

    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        pass

    def check_custom_dependencies(self): pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """
        return ['preproc_dwi', 'preproc_bvec', 'preproc_bval', 'b0_mask']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """
        output_list_dti = [
            'dti', 'fa', 'md', 'ad', 'rd', 'decfa',
            'registered_fa', 'registered_md', 'registered_ad', 'registered_rd',
            'statistics_fa', 'statistics_md', 'statistics_ad', 'statistics_rd',
            'b_spline_transform', 'affine_matrix'
        ]

        return output_list_dti

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        import clinica.utils.input_files as input_files
        from clinica.utils.stream import cprint

        all_errors = []

        # b0 Mask
        try:
            b0_mask = clinica_file_reader(self.subjects,
                                          self.sessions,
                                          self.caps_directory,
                                          input_files.DWI_PREPROC_BRAINMASK)
        except ClinicaException as e:
            all_errors.append(e)

        # DWI preprocessing NIfTI
        try:
            dwi_caps = clinica_file_reader(self.subjects,
                                           self.sessions,
                                           self.caps_directory,
                                           input_files.DWI_PREPROC_NII)
        except ClinicaException as e:
            all_errors.append(e)

        # bval files
        try:
            bval_files = clinica_file_reader(self.subjects,
                                             self.sessions,
                                             self.caps_directory,
                                             input_files.DWI_PREPROC_BVAL)
        except ClinicaException as e:
            all_errors.append(e)

        # bvec files
        try:
            bvec_files = clinica_file_reader(self.subjects,
                                             self.sessions,
                                             self.caps_directory,
                                             input_files.DWI_PREPROC_BVEC)
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = 'Clinica faced error(s) while trying to read files in your CAPS directory.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaCAPSError(error_message)

        read_input_node = npe.Node(name="LoadingCLIArguments",
                                   interface=nutil.IdentityInterface(
                                       fields=self.get_input_fields(),
                                       mandatory_inputs=True),
                                   iterables=[('preproc_dwi', dwi_caps),
                                              ('preproc_bvec', bvec_files),
                                              ('preproc_bval', bval_files),
                                              ('b0_mask', b0_mask)],
                                   synchronize=True)

        self.connect([
            (read_input_node, self.input_node, [('b0_mask', 'b0_mask')]),
            (read_input_node, self.input_node, [('preproc_dwi', 'preproc_dwi')]),
            (read_input_node, self.input_node, [('preproc_bval', 'preproc_bval')]),
            (read_input_node, self.input_node, [('preproc_bvec', 'preproc_bvec')])
        ])
        cprint('The pipeline will last approximately 20 minutes per image.')

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        from clinica.utils.nipype import fix_join

        import clinica.pipelines.dwi_dti.dwi_dti_utils as utils

        # Find container path from filename
        container_path = npe.Node(nutil.Function(
            input_names=['dwi_filename'],
            output_names=['container'],
            function=utils.dwi_container_from_filename),
            name='container_path')

        rename_into_caps = npe.Node(nutil.Function(
            input_names=['in_caps_dwi', 'in_norm_fa', 'in_norm_md',
                         'in_norm_ad', 'in_norm_rd', 'in_b_spline_transform',
                         'in_affine_matrix'],
            output_names=['out_caps_fa', 'out_caps_md', 'out_caps_ad',
                          'out_caps_rd', 'out_caps_b_spline_transform',
                          'out_caps_affine_matrix'],
            function=utils.rename_into_caps),
            name='rename_into_caps')

        # Writing results into CAPS
        write_results = npe.Node(name='write_results',
                                 interface=nio.DataSink())
        write_results.inputs.base_directory = self.caps_directory
        write_results.inputs.parameterization = False

        self.connect([
           (self.input_node, container_path, [('preproc_dwi', 'dwi_filename')]),  # noqa

           (container_path,   write_results, [(('container', fix_join, 'dwi', 'dti_based_processing'), 'container')]),  # noqa
           (self.output_node, write_results, [('dti',   'native_space.@dti')]),  # noqa
           (self.output_node, write_results, [('fa',    'native_space.@fa'),  # noqa
                                              ('md',    'native_space.@md'),  # noqa
                                              ('ad',    'native_space.@ad'),  # noqa
                                              ('rd',    'native_space.@rd'),  # noqa
                                              ('decfa', 'native_space.@decfa')]),  # noqa

           (self.input_node,  rename_into_caps, [('preproc_dwi',        'in_caps_dwi')]),  # noqa
           (self.output_node, rename_into_caps, [('registered_fa',      'in_norm_fa'),  # noqa
                                                 ('registered_md',      'in_norm_md'),  # noqa
                                                 ('registered_ad',      'in_norm_ad'),  # noqa
                                                 ('registered_rd',      'in_norm_rd'),  # noqa
                                                 ('affine_matrix',      'in_affine_matrix'),  # noqa
                                                 ('b_spline_transform', 'in_b_spline_transform')]),  # noqa

           (rename_into_caps, write_results, [('out_caps_fa',                 'normalized_space.@registered_fa'),  # noqa
                                              ('out_caps_md',                 'normalized_space.@registered_md'),  # noqa
                                              ('out_caps_ad',                 'normalized_space.@registered_ad'),  # noqa
                                              ('out_caps_rd',                 'normalized_space.@registered_rd'),  # noqa
                                              ('out_caps_affine_matrix',      'normalized_space.@affine_matrix'),  # noqa
                                              ('out_caps_b_spline_transform', 'normalized_space.@b_spline_transform')]),  # noqa

           (self.output_node, write_results, [('statistics_fa', 'atlas_statistics.@statistics_fa'),  # noqa
                                              ('statistics_md', 'atlas_statistics.@statistics_md'),  # noqa
                                              ('statistics_ad', 'atlas_statistics.@statistics_ad'),  # noqa
                                              ('statistics_rd', 'atlas_statistics.@statistics_rd')])   # noqa
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """
        import clinica.pipelines.dwi_dti.dwi_dti_workflows as workflows
        import clinica.pipelines.dwi_dti.dwi_dti_utils as utils

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix as mrtrix
        from clinica.lib.nipype.interfaces.mrtrix3.utils import TensorMetrics
        from clinica.lib.nipype.interfaces.mrtrix.preprocess import DWI2Tensor

        # Nodes creation
        # ==============
        get_bids_identifier = npe.Node(interface=nutil.Function(
            input_names=['caps_dwi_filename'],
            output_names=['bids_identifier'],
            function=utils.extract_bids_identifier_from_caps_filename),
            name='0-Get_BIDS_Identifier')

        get_caps_filenames = npe.Node(interface=nutil.Function(
            input_names=['caps_dwi_filename'],
            output_names=['bids_source', 'out_dti',
                          'out_fa', 'out_md', 'out_ad', 'out_rd'],
            function=utils.get_caps_filenames),
            name='0-CAPS_Filenames')

        convert_gradients = npe.Node(
            interface=mrtrix.FSL2MRTrix(),
            name='0-Convert_FSL_Gradient')

        dwi_to_dti = npe.Node(
            # interface=mrtrix.DWI2Tensor(),
            interface=DWI2Tensor(),
            name='1-Compute_DTI')

        dti_to_metrics = npe.Node(
            #            interface=mrtrix3.TensorMetrics(),
            interface=TensorMetrics(),
            name='2-DTI-based_Metrics')

        register_on_jhu_atlas = workflows.register_dti_maps_on_atlas(
            working_directory=self.base_dir,
            name="3-Register_DTI_Maps_On_JHU")

        scalar_analysis = npe.Node(
            interface=nutil.Function(
                input_names=['in_registered_map', 'name_map', 'prefix_file'],
                output_names=['atlas_statistics_list'],
                function=utils.statistics_on_atlases),
            name='4-Scalar_Analysis')
        scalar_analysis_fa = scalar_analysis.clone('4-Scalar_Analysis_FA')
        scalar_analysis_fa.inputs.name_map = 'FA'
        scalar_analysis_md = scalar_analysis.clone('4-Scalar_Analysis_MD')
        scalar_analysis_md.inputs.name_map = 'MD'
        scalar_analysis_ad = scalar_analysis.clone('4-Scalar_Analysis_AD')
        scalar_analysis_ad.inputs.name_map = 'AD'
        scalar_analysis_rd = scalar_analysis.clone('4-Scalar_Analysis_RD')
        scalar_analysis_rd.inputs.name_map = 'RD'

        thres_map = npe.Node(fsl.Threshold(thresh=0.0),
                             iterfield=['in_file'],
                             name='5-Remove_Negative')
        thres_fa = thres_map.clone('5-Remove_Negative_FA')
        thres_md = thres_map.clone('5-Remove_Negative_MD')
        thres_ad = thres_map.clone('5-Remove_Negative_AD')
        thres_rd = thres_map.clone('5-Remove_Negative_RD')

        print_begin_message = npe.Node(
            interface=nutil.Function(
                input_names=['in_bids_or_caps_file'],
                function=utils.print_begin_pipeline),
            name='Write-Begin_Message')

        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=['in_bids_or_caps_file', 'final_file_1', 'final_file_2'],
                function=utils.print_end_pipeline),
            name='Write-End_Message')

        # Connection
        # ==========
        self.connect([
            (self.input_node, get_caps_filenames, [('preproc_dwi', 'caps_dwi_filename')]),  # noqa
            # Print begin message
            (self.input_node, print_begin_message, [('preproc_dwi', 'in_bids_or_caps_file')]),  # noqa
            # Get BIDS/CAPS identifier from filename
            (self.input_node, get_bids_identifier, [('preproc_dwi', 'caps_dwi_filename')]),  # noqa
            # Convert FSL gradient files (bval/bvec) to MRtrix format
            (self.input_node, convert_gradients, [('preproc_bval', 'bval_file'),  # noqa
                                                  ('preproc_bvec', 'bvec_file')]),  # noqa
            # Computation of the DTI model
            (self.input_node,    dwi_to_dti, [('b0_mask',       'in_mask'),  # noqa
                                              ('preproc_dwi',   'in_file')]),  # noqa
            (convert_gradients,  dwi_to_dti, [('encoding_file', 'encoding_file')]),  # noqa
            (get_caps_filenames, dwi_to_dti, [('out_dti',       'out_filename')]),  # noqa
            # Computation of the different metrics from the DTI
            (get_caps_filenames, dti_to_metrics, [('out_fa',  'out_fa')]),  # noqa
            (get_caps_filenames, dti_to_metrics, [('out_md',  'out_adc')]),  # noqa
            (get_caps_filenames, dti_to_metrics, [('out_ad',  'out_ad')]),  # noqa
            (get_caps_filenames, dti_to_metrics, [('out_rd',  'out_rd')]),  # noqa
            (self.input_node,    dti_to_metrics, [('b0_mask', 'in_mask')]),  # noqa
            (dwi_to_dti,         dti_to_metrics, [('tensor',  'in_file')]),  # noqa
            # Register DTI maps on JHU atlas
            (dti_to_metrics,    register_on_jhu_atlas, [('out_fa',  'inputnode.in_fa'),  # noqa
                                                        ('out_adc', 'inputnode.in_md'),  # noqa
                                                        ('out_ad',  'inputnode.in_ad'),  # noqa
                                                        ('out_rd',  'inputnode.in_rd')]),  # noqa
            # Generate regional TSV files
            (get_bids_identifier,   scalar_analysis_fa, [('bids_identifier',        'prefix_file')]),  # noqa
            (register_on_jhu_atlas, scalar_analysis_fa, [('outputnode.out_norm_fa', 'in_registered_map')]),  # noqa
            (get_bids_identifier,   scalar_analysis_md, [('bids_identifier',        'prefix_file')]),  # noqa
            (register_on_jhu_atlas, scalar_analysis_md, [('outputnode.out_norm_md', 'in_registered_map')]),  # noqa
            (get_bids_identifier,   scalar_analysis_ad, [('bids_identifier',        'prefix_file')]),  # noqa
            (register_on_jhu_atlas, scalar_analysis_ad, [('outputnode.out_norm_ad', 'in_registered_map')]),  # noqa
            (get_bids_identifier,   scalar_analysis_rd, [('bids_identifier',        'prefix_file')]),  # noqa
            (register_on_jhu_atlas, scalar_analysis_rd, [('outputnode.out_norm_rd', 'in_registered_map')]),  # noqa
            # Remove negative values from the DTI maps:
            (get_caps_filenames, thres_fa, [('out_fa',  'out_file')]),  # noqa
            (dti_to_metrics,     thres_fa, [('out_fa',  'in_file')]),  # noqa

            (get_caps_filenames, thres_md, [('out_md',  'out_file')]),  # noqa
            (dti_to_metrics,     thres_md, [('out_adc', 'in_file')]),  # noqa

            (get_caps_filenames, thres_ad, [('out_ad',  'out_file')]),  # noqa
            (dti_to_metrics,     thres_ad, [('out_ad',  'in_file')]),  # noqa

            (get_caps_filenames, thres_rd, [('out_rd',  'out_file')]),  # noqa
            (dti_to_metrics,     thres_rd, [('out_rd',  'in_file')]),  # noqa
            # Outputnode
            (dwi_to_dti,            self.output_node, [('tensor',   'dti')]),  # noqa
            (thres_fa,              self.output_node, [('out_file', 'fa')]),  # noqa
            (thres_md,              self.output_node, [('out_file', 'md')]),  # noqa
            (thres_ad,              self.output_node, [('out_file', 'ad')]),  # noqa
            (thres_rd,              self.output_node, [('out_file', 'rd')]),  # noqa
            (dti_to_metrics,        self.output_node, [('out_evec', 'decfa')]),  # noqa
            (register_on_jhu_atlas, self.output_node, [('outputnode.out_norm_fa',            'registered_fa'),  # noqa
                                                       ('outputnode.out_norm_md',            'registered_md'),  # noqa
                                                       ('outputnode.out_norm_ad',            'registered_ad'),  # noqa
                                                       ('outputnode.out_norm_rd',            'registered_rd'),  # noqa
                                                       ('outputnode.out_affine_matrix',      'affine_matrix'),  # noqa
                                                       ('outputnode.out_b_spline_transform', 'b_spline_transform')]),  # noqa
            (scalar_analysis_fa,    self.output_node, [('atlas_statistics_list', 'statistics_fa')]),  # noqa
            (scalar_analysis_md,    self.output_node, [('atlas_statistics_list', 'statistics_md')]),  # noqa
            (scalar_analysis_ad,    self.output_node, [('atlas_statistics_list', 'statistics_ad')]),  # noqa
            (scalar_analysis_rd,    self.output_node, [('atlas_statistics_list', 'statistics_rd')]),   # noqa
            # Print end message
            (self.input_node,    print_end_message, [('preproc_dwi',           'in_bids_or_caps_file')]),  # noqa
            (thres_rd,           print_end_message, [('out_file',              'final_file_1')]),  # noqa
            (scalar_analysis_rd, print_end_message, [('atlas_statistics_list', 'final_file_2')]),  # noqa
        ])



# coding: utf-8


def visits_to_timepoints(subject, mri_list_subj, adnimerge_subj, modality, visit_field="VISIT",
                         scandate_field="SCANDATE"):
    """
    A correspondence is established between ADNIMERGE and MRILIST visits, since they have different notations.

    Args:
        subject: Subject identifier
        mri_list_subj: Dataframe containing list of MRI scans for the subject
        adnimerge_subj: Dataframe containing visits data for the subject
        modality: Imaging modality

    Returns:

    """
    from clinica.utils.stream import cprint

    if modality == 'T1':
        mri_list_subj = mri_list_subj[mri_list_subj[visit_field] != 'ADNI Baseline']

    visits = dict()
    unique_visits = list(mri_list_subj[visit_field].unique())
    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():
        visit = adni_row[1]

        preferred_visit_name = get_preferred_visit_name(visit)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits.keys():
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint('[%s] Subject %s has multiple visits for one timepoint.' % (modality, subject))

            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:

        image = (mri_list_subj[mri_list_subj[visit_field] == visit]).iloc[0]

        key_min_visit = get_closest_visit(image, pending_timepoints, subject, visit_field, scandate_field)

        if not key_min_visit:
            continue

        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image[visit_field]
        elif visits[key_min_visit] != image[visit_field]:
            cprint('[%s] Subject %s has multiple visits for one timepoint.' % (modality, subject))

    return visits


def get_preferred_visit_name(visit):
    """
    Provides the expected visit name for a given visit
    Args:
        visit: A visit entry from ADNIMERGE

    Returns:
        string: expected visit name

    """

    if visit.ORIGPROT == 'ADNI3':
        if visit.VISCODE == 'bl':
            preferred_visit_name = 'ADNI Screening'
        else:
            year = str(int(visit.VISCODE[1:]) / 12)
            preferred_visit_name = 'ADNI3 Year ' + year + ' Visit'
    elif visit.ORIGPROT == 'ADNI2':
        if visit.VISCODE == 'bl':
            preferred_visit_name = 'ADNI2 Screening MRI-New Pt'
        elif visit.VISCODE == 'm03':
            preferred_visit_name = 'ADNI2 Month 3 MRI-New Pt'
        elif visit.VISCODE == 'm06':
            preferred_visit_name = 'ADNI2 Month 6-New Pt'
        else:
            year = str(int(visit.VISCODE[1:]) / 12)
            preferred_visit_name = 'ADNI2 Year ' + year + ' Visit'
    else:
        if visit.VISCODE == 'bl':
            if visit.ORIGPROT == 'ADNI1':
                preferred_visit_name = 'ADNI Screening'
            else:  # ADNIGO
                preferred_visit_name = 'ADNIGO Screening MRI'
        elif visit.VISCODE == 'm03':  # Only for ADNIGO Month 3
            preferred_visit_name = 'ADNIGO Month 3 MRI'
        else:
            month = int(visit.VISCODE[1:])
            if month < 54:
                preferred_visit_name = 'ADNI1/GO Month ' + str(month)
            else:
                preferred_visit_name = 'ADNIGO Month ' + str(month)
    return preferred_visit_name


def get_closest_visit(image, pending_timepoints, subject, visit_field, scandate_field):
    """
    Chooses the visit with the closest date to a given image acquisition date

    Args:
        image: An image entry from MPRAGEMETA
        pending_timepoints: List of visit entries from ADNIMERGE
        subject: Subject identifier
        visit_field:
        scandate_field:

    Returns:

    """

    from datetime import datetime
    from clinica.utils.stream import cprint

    min_db = 100000
    min_db2 = 0
    min_visit = None
    min_visit2 = None

    for timepoint in pending_timepoints:
        db = days_between(image[scandate_field], timepoint.EXAMDATE)
        if db < min_db:
            min_db2 = min_db
            min_visit2 = min_visit

            min_db = db
            min_visit = timepoint

    if min_visit is None:
        cprint('No corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' + image[visit_field])
        cprint(image)
        return None

    if min_visit2 is not None and min_db > 90:
        cprint('More than 60 days for corresponding timepoint in ADNIMERGE for subject ' + subject + ' in visit ' +
               image[visit_field] + ' on ' + image[scandate_field])
        cprint('Timepoint 1: ' + min_visit.VISCODE + ' - ' + min_visit.ORIGPROT + ' on ' + min_visit.EXAMDATE +
               ' (Distance: ' + str(min_db) + ' days)')
        cprint('Timepoint 2: ' + min_visit2.VISCODE + ' - ' + min_visit2.ORIGPROT + ' on ' + min_visit2.EXAMDATE +
               ' (Distance: ' + str(min_db2) + ' days)')

        # If image is too close to the date between two visits we prefer the earlier visit
        if (datetime.strptime(min_visit.EXAMDATE, "%Y-%m-%d")
                > datetime.strptime(image[scandate_field], "%Y-%m-%d")
                > datetime.strptime(min_visit2.EXAMDATE, "%Y-%m-%d")):
            dif = days_between(min_visit.EXAMDATE, min_visit2.EXAMDATE)
            if abs((dif / 2.0) - min_db) < 30:
                min_visit = min_visit2

        cprint('We prefer ' + min_visit.VISCODE)

    key_min_visit = (min_visit.VISCODE, min_visit.COLPROT, min_visit.ORIGPROT)

    return key_min_visit


def days_between(d1, d2):
    """
    Calculate the days between two dates

    Args:
        d1: date 1
        d2: date 2

    Returns: number of days between date 2 and date 1

    """

    from datetime import datetime
    d1 = datetime.strptime(d1, "%Y-%m-%d")
    d2 = datetime.strptime(d2, "%Y-%m-%d")
    return abs((d2 - d1).days)


def select_image_qc(id_list, mri_qc_subj):
    """
    An image is selected from several scans according to QC

    Args:
        id_list: List of images identifiers to choose among
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: int - Chosen image identifier

    """

    import numpy as np

    if len(id_list) == 0:
        return None

    selected_image = None
    image_ids = ['I' + str(imageuid) for imageuid in id_list]
    int_ids = [int(imageuid) for imageuid in id_list]
    images_qc = mri_qc_subj[mri_qc_subj.loni_image.isin(image_ids)]

    if images_qc.empty:
        return max(int_ids)

    if np.sum(images_qc.series_selected) == 1:
        selected_image = images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
    else:
        images_not_rejected = images_qc[images_qc.series_quality < 4]

        if images_not_rejected.empty:

            # There are no images that passed the qc
            # so we'll try to see if there are other images without qc,
            # otherwise return None
            qc_ids = set([int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()])
            no_qc_ids = list(set(int_ids) - qc_ids)

            if len(no_qc_ids) == 0:
                return None
            else:
                return max(no_qc_ids)

        # We select the image with the best (lower) QC.
        # If no positive QC available we choose image with -1 (not performed)
        series_quality = [q if q > 0 else 4 for q in list(images_not_rejected.series_quality)]
        best_q = np.amin(series_quality)
        if best_q == 4:
            best_q = -1
        images_best_qc = images_not_rejected[images_not_rejected.series_quality == best_q]
        if len(images_best_qc) == 1:
            selected_image = images_best_qc.iloc[0].loni_image[1:]
        else:
            best_ids = [int(x[1:]) for x in images_best_qc.loni_image.unique()]
            selected_image = max(best_ids)

    return int(selected_image)


def get_images_pet(subject, pet_qc_subj, subject_pet_meta, df_cols, modality, sequences_preprocessing_step,
                   viscode_field="VISCODE2"):
    """
    Selection of scans passing QC and at the chosen preprocessing stage is performed

    Args:
        subject: Subject identifier
        pet_qc_subj: Dataframe containing QC for scans for the subject
        subject_pet_meta: Dataframe containing metadata for scans for the subject
        df_cols: Columns of output dataframe
        modality: Imaging modality
        sequences_preprocessing_step: List of sequence names that correspond to desired preprocessing stage
        viscode_field: Name of the field in the pet_qc_subj dataframe that provides to the visit code

    Returns: Dataframe containing images metadata

    """

    import pandas as pd
    from clinica.utils.stream import cprint

    subj_dfs_list = []

    for visit in list(pet_qc_subj[viscode_field].unique()):
        if pd.isna(visit):
            continue

        pet_qc_visit = pet_qc_subj[pet_qc_subj[viscode_field] == visit]

        if pet_qc_visit.empty:
            continue

        # If there are several scans for a timepoint we start with image acquired last (higher LONIUID)
        pet_qc_visit = pet_qc_visit.sort_values("LONIUID", ascending=False)

        original_pet_meta = pd.DataFrame(columns=subject_pet_meta.columns)
        qc_visit = pet_qc_visit.iloc[0]
        for qc_index in range(len(pet_qc_visit)):
            qc_visit = pet_qc_visit.iloc[qc_index]

            # We are looking for FDG PET metadata of Original images, that passed QC,
            # acquired at the same date as the current scan that passed QC for the current visit,
            # not containing ‘early’ in the sequence name

            original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                 & (subject_pet_meta['Image ID'] == int(qc_visit.LONIUID[1:]))
                                                 & (subject_pet_meta['Scan Date'] == qc_visit.EXAMDATE)
                                                 & ~subject_pet_meta.Sequence.str.contains("early",
                                                                                           case=False, na=False)]
            # Check if we found a matching image. If yes, we stop looking for it.
            if not original_pet_meta.empty:
                break

        if original_pet_meta.empty:
            cprint('No %s images metadata for subject %s and visit %s' % (modality, subject, qc_visit[viscode_field]))
            continue

        original_image = original_pet_meta.iloc[0]

        # Co-registered and Averaged image with the same Series ID of the original image
        averaged_pet_meta = subject_pet_meta[subject_pet_meta['Sequence'].isin(sequences_preprocessing_step)
                                             & (subject_pet_meta['Series ID'] == original_image['Series ID'])]

        # If an explicit Co-registered, Averaged image does not exist,
        # the original image is already in that preprocessing stage.

        if averaged_pet_meta.empty:
            sel_image = original_image
            original = True
        else:
            sel_image = averaged_pet_meta.iloc[0]
            original = False

        phase = 'ADNI1' if modality == 'PIB-PET' else qc_visit.Phase
        visit = sel_image.Visit
        sequence = replace_sequence_chars(sel_image.Sequence)
        date = sel_image['Scan Date']
        study_id = sel_image['Study ID']
        series_id = sel_image['Series ID']
        image_id = sel_image['Image ID']

        # If it is an amyloid PET we need to find which is the tracer of the scan and add it to the
        if modality == 'Amyloid-PET':
            if 'av45' in sel_image.Sequence.lower():
                tracer = 'AV45'
            elif 'fbb' in sel_image.Sequence.lower():
                tracer = 'FBB'
            else:
                cprint('Unknown tracer for Amyloid PET image metadata for subject ' + subject +
                       ' for visit ' + qc_visit[viscode_field])
                continue

            scan_data = [[phase, subject, qc_visit[viscode_field], str(visit), sequence, date, str(study_id),
                          str(series_id), str(image_id), original, tracer]]
        else:
            scan_data = [[phase, subject, qc_visit[viscode_field], str(visit), sequence, date, str(study_id),
                          str(series_id), str(image_id), original]]

        row_to_append = pd.DataFrame(scan_data, columns=df_cols)
        subj_dfs_list.append(row_to_append)

    return subj_dfs_list


def replace_sequence_chars(sequence_name):
    """
    Replace special characters in the sequence by underscores (as done for corresponding folder names in ADNI)

    Args:
        sequence_name: sequence to process

    Returns: the new string

    """
    import re
    return re.sub('[ /;*()<>:]', '_', sequence_name)


def write_adni_sessions_tsv(sessions_dict, fields_bids, bids_subjs_paths):
    """
    Write the result of method create_session_dict into several tsv files

    Args:
        sessions_dict: dictonary coming from the method create_sessions_dict
        fields_bids: fields bids to convert
        bids_subjs_paths: a list with the path to all bids subjects

    Returns:

    """
    from os import path
    import os
    import pandas as pd
    import numpy as np

    columns_order = remove_fields_duplicated(fields_bids)

    columns_order.insert(0, 'session_id')
    for sp in bids_subjs_paths:

        if not path.exists(sp):
            os.makedirs(sp)

        bids_id = sp.split(os.sep)[-1]

        fields_bids = list(set(fields_bids))
        sessions_df = pd.DataFrame(columns=fields_bids)

        if bids_id in sessions_dict:
            sess_aval = sessions_dict[bids_id].keys()
            for ses in sess_aval:
                sessions_df = sessions_df.append(pd.DataFrame(sessions_dict[bids_id][ses], index=['i', ]))

            sessions_df = sessions_df[columns_order]

            sessions_df[[
                'adas_Q1', 'adas_Q2', 'adas_Q3', 'adas_Q4',
                'adas_Q5', 'adas_Q6', 'adas_Q7', 'adas_Q8',
                'adas_Q9', 'adas_Q10', 'adas_Q11', 'adas_Q12',
                'adas_Q13'
                ]] = sessions_df[[
                    'adas_Q1', 'adas_Q2', 'adas_Q3', 'adas_Q4',
                    'adas_Q5', 'adas_Q6', 'adas_Q7', 'adas_Q8',
                    'adas_Q9', 'adas_Q10', 'adas_Q11', 'adas_Q12',
                    'adas_Q13']].apply(pd.to_numeric)

            sessions_df['adas_memory'] = (sessions_df['adas_Q1']
                                          + sessions_df['adas_Q4']
                                          + sessions_df['adas_Q7']
                                          + sessions_df['adas_Q8']
                                          + sessions_df['adas_Q9'])  # / 45
            sessions_df['adas_language'] = (sessions_df['adas_Q2']
                                            + sessions_df['adas_Q5']
                                            + sessions_df['adas_Q10']
                                            + sessions_df['adas_Q11']
                                            + sessions_df['adas_Q12'])  # / 25
            sessions_df['adas_praxis'] = (sessions_df['adas_Q3']
                                          + sessions_df['adas_Q6'])  # / 10
            sessions_df['adas_concentration'] = sessions_df['adas_Q13']  # / 5

            sessions_df.to_csv(path.join(sp, bids_id + '_sessions.tsv'), sep='\t', index=False, encoding='utf-8')


def remove_fields_duplicated(bids_fields):
    """
    Remove duplicated fields in a list

    Args:
        bids_fields: List of fields

    Returns: list

    """
    seen = set()
    seen_add = seen.add
    return [x for x in bids_fields if not (x in seen or seen_add(x))]


def create_adni_sessions_dict(bids_ids, clinic_specs_path, clinical_data_dir, bids_subjs_paths):
    """
    Extract all the data required for the sessions files and organize them in a
    dictionary

    Args: bids_ids: clinic_specs_path: path to the specifications for
    converting the clinical data clinical_data_dir: path to the clinical data
    folder bids_subjs_paths: a list with the path to all the BIDS subjects

    Returns:

    """
    import pandas as pd
    from os import path
    from datetime import datetime
    import clinica.iotools.bids_utils as bids
    from clinica.utils.stream import cprint

    # Load data
    sessions = pd.read_excel(clinic_specs_path, sheet_name='sessions.tsv')
    sessions_fields = sessions['ADNI']
    field_location = sessions['ADNI location']
    sessions_fields_bids = sessions['BIDS CLINICA']
    fields_dataset = []
    previous_location = ''
    fields_bids = []
    sessions_dict = {}

    for i in range(0, len(sessions_fields)):
        if not pd.isnull(sessions_fields[i]):
            fields_bids.append(sessions_fields_bids[i])
            fields_dataset.append(sessions_fields[i])

    sessions_df = pd.DataFrame(columns=fields_bids)

    for i in range(0, len(field_location)):
        # If the i-th field is available
        if (not pd.isnull(field_location[i])) and path.exists(path.join(clinical_data_dir,
                                                                        field_location[i].split('/')[0])):
            # Load the file
            tmp = field_location[i].split('/')
            location = tmp[0]
            if location != previous_location:
                previous_location = location
                file_to_read_path = path.join(clinical_data_dir, location)
                cprint('\tReading clinical data file : ' + location)
                file_to_read = pd.read_csv(file_to_read_path, dtype=str)

                for r in range(0, len(file_to_read.values)):
                    row = file_to_read.iloc[r]

                    # Depending of the file that needs to be open, identify and
                    # do needed preprocessing on the column that contains the
                    # subjects ids
                    if location == 'ADNIMERGE.csv':
                        id_ref = 'PTID'
                        # What was the meaning of this line ?
                        # subj_id = row[id_ref.decode('utf-8')]
                        subj_id = row[id_ref]
                        subj_id = bids.remove_space_and_symbols(subj_id)
                    else:
                        id_ref = 'RID'
                        rid = str(row[id_ref])

                        # Fill the rid with the needed number of zero
                        if 4 - len(rid) > 0:
                            zeros_to_add = 4 - len(rid)
                            subj_id = '0' * zeros_to_add + rid
                        else:
                            subj_id = rid

                    # Extract the BIDS subject id related with the original
                    # subject id
                    subj_bids = [s for s in bids_ids if subj_id in s]

                    if len(subj_bids) == 0:
                        pass
                    elif len(subj_bids) > 1:
                        raise('Error: multiple subjects found for the same RID')
                    else:
                        subj_bids = subj_bids[0]
                        for j in range(0, len(sessions_fields)):
                            # If the i-th field is available
                            if not pd.isnull(sessions_fields[j]):
                                # Extract only the fields related to the current file opened
                                if location in field_location[i]:
                                    if location in ['ADAS_ADNIGO2.csv', 'DXSUM_PDXCONV_ADNIALL.csv', 'CDR.csv',
                                                    'NEUROBAT.csv', 'GDSCALE.csv', 'MODHACH.csv', 'MOCA.csv',
                                                    'NPIQ.csv', 'MEDHIST.csv', 'VITALS.csv', 'UWNPSYCHSUM_03_07_19.csv',
                                                    'ECOGPT.csv', 'ECOGSP.csv', 'FCI.csv', 'CCI.csv']:
                                        if pd.isnull(row['VISCODE2']) or row['VISCODE2'] == 'f':
                                            continue
                                        visit_id = row['VISCODE2']
                                        # Convert sc to bl
                                        if visit_id == 'sc':
                                            visit_id = 'bl'
                                    elif location in ['BHR_EVERYDAY_COGNITION.csv', 'BHR_BASELINE_QUESTIONNAIRE.csv',
                                                      'BHR_LONGITUDINAL_QUESTIONNAIRE.csv']:
                                        visit_id = row['Timepoint']
                                    else:
                                        visit_id = row['VISCODE']
                                    try:
                                        field_value = row[sessions_fields[j]]
                                        bids_field_name = sessions_fields_bids[j]

                                        # Calculating age from ADNIMERGE
                                        if (sessions_fields[j] == "AGE") and (visit_id != "bl"):
                                            examdate = datetime.strptime(row["EXAMDATE"], "%Y-%m-%d")
                                            examdate_bl = datetime.strptime(row["EXAMDATE_bl"], "%Y-%m-%d")
                                            delta = examdate - examdate_bl

                                            # Adding time passed since bl to patient's age in current visit
                                            field_value = round(float(field_value) + (delta.days / 365.25), 1)

                                        sessions_dict = update_sessions_dict(sessions_dict, subj_bids, visit_id,
                                                                             field_value, bids_field_name)
                                    except KeyError:
                                        pass
                                        # cprint('Field value ' + Fore.RED + sessions_fields[j] + Fore.RESET +
                                        # ' could not be added to sessions.tsv')
            else:
                continue

    # Write the sessions dictionary created in several tsv files
    write_adni_sessions_tsv(sessions_dict, fields_bids, bids_subjs_paths)


def update_sessions_dict(sessions_dict, subj_bids, visit_id, field_value, bids_field_name):
    """
    Update the sessions dictionary for the bids subject specified by subj_bids
    created by the method create_adni_sessions_dict

    Args:
        sessions_dict: the session_dict created by the method create_adni_sessions_dict
        subj_bids: bids is of the subject for which the information need to be updated
        visit_id: session name (ex m54)
        field_value: value of the field extracted from the original clinical data
        bids_field_name: BIDS name of the field to update (Ex: diagnosis or examination date)

    Returns:
        session_dict: the dictonary updated

    """

    from pandas import isna

    if visit_id == 'sc' or visit_id == 'uns1':
        return sessions_dict

    visit_id = viscode_to_session(visit_id)

    if bids_field_name == 'diagnosis':
        field_value = convert_diagnosis_code(field_value)

    # If the dictionary already contain the subject add or update information
    # regarding a specific session, otherwise create the entry
    if subj_bids in sessions_dict:
        sess_available = sessions_dict[subj_bids].keys()

        if visit_id in sess_available:
            # If a value is already contained, update it only if the previous value is nan
            if bids_field_name in sessions_dict[subj_bids][visit_id]:

                if isna(sessions_dict[subj_bids][visit_id][bids_field_name]):
                    sessions_dict[subj_bids][visit_id].update(
                        {bids_field_name: field_value})
            else:
                sessions_dict[subj_bids][visit_id].update(
                    {bids_field_name: field_value})
        else:
            sessions_dict[subj_bids].update({visit_id: {'session_id': 'ses-' + visit_id,
                                                        bids_field_name: field_value}})
    else:
        sessions_dict.update({subj_bids: {visit_id: {'session_id': 'ses-' + visit_id,
                                                     bids_field_name: field_value}}})

    return sessions_dict


def convert_diagnosis_code(diagnosis_code):
    """
    Convert the string field DX contained in ADNIMERGE.csv into a code that identify the diagnosis

    Args:
        diagnosis_code: a string that represents the diagnosis

    Returns:
        A code that identify a diagnosis

    """

    from pandas import isna

    # Legend
    diagnosis = {'CN': 'CN', 'MCI': 'MCI', 'Dementia': 'AD'}

    if isna(diagnosis_code):
        return diagnosis_code
    else:
        return diagnosis[diagnosis_code]


def create_adni_scans_files(clinic_specs_path, bids_subjs_paths, bids_ids):
    """
    Create scans.tsv files for ADNI

    Args:
        clinic_specs_path: path to the clinical file
        bids_subjs_paths: list of bids subject paths
        bids_ids: list of bids ids

    Returns:

    """
    from glob import glob
    import os
    import pandas as pd
    from os import path
    from os.path import normpath

    scans_dict = {}

    for bids_id in bids_ids:
        scans_dict.update({bids_id: {'T1/DWI/fMRI': {}, 'FDG': {}}})

    scans_specs = pd.read_excel(clinic_specs_path, sheet_name='scans.tsv')
    scans_fields_db = scans_specs['ADNI']
    scans_fields_bids = scans_specs['BIDS CLINICA']
    scans_fields_mod = scans_specs['Modalities related']
    fields_bids = ['filename']

    for i in range(0, len(scans_fields_db)):
        if not pd.isnull(scans_fields_db[i]):
            fields_bids.append(scans_fields_bids[i])

    scans_df = pd.DataFrame(columns=(fields_bids))

    for bids_subj_path in bids_subjs_paths:
        # Create the file
        bids_id = os.path.basename(normpath(bids_subj_path))

        sessions_paths = glob(path.join(bids_subj_path, 'ses-*'))
        for session_path in sessions_paths:
            session_name = session_path.split(os.sep)[-1]
            tsv_name = bids_id + '_' + session_name + "_scans.tsv"

            # If the file already exists, remove it
            if os.path.exists(path.join(session_path, tsv_name)):
                os.remove(path.join(session_path, tsv_name))

            scans_tsv = open(path.join(session_path, tsv_name), 'a')
            scans_df.to_csv(scans_tsv, sep='\t', index=False, encoding='utf-8')

            # Extract modalities available for each subject
            mod_available = glob(path.join(session_path, '*'))
            for mod in mod_available:
                mod_name = os.path.basename(mod)
                files = glob(path.join(mod, '*'))
                for file in files:
                    file_name = os.path.basename(file)
                    if mod == "anat" or mod == "dwi" or mod == "func":
                        type_mod = 'T1/DWI/fMRI'
                    else:
                        type_mod = 'FDG'

                    scans_df['filename'] = pd.Series(path.join(mod_name, file_name))
                    scans_df.to_csv(scans_tsv, header=False, sep='\t', index=False, encoding='utf-8')

            scans_df = pd.DataFrame(columns=(fields_bids))


def find_image_path(images, source_dir, modality, prefix, id_field):
    """
    For each image, the path to an existing image file or folder is created from image metadata.

    Args:
        images: List of images metadata
        source_dir: path to the ADNI directory
        modality: Imaging modality
        prefix: Prefix to prepend to image identifier to create name of folder containing the image
        id_field: Name of the field in images metadata dataframe containing the image identifier to use

    Returns: Dataframe containing metadata and existing paths

    """

    from os import path, walk
    import pandas as pd
    from clinica.utils.stream import cprint

    is_dicom = []
    image_folders = []

    for row in images.iterrows():
        image = row[1]
        seq_path = path.join(source_dir, str(image.Subject_ID), image.Sequence)
        image_path = ''
        for (dirpath, dirnames, filenames) in walk(seq_path):
            found = False
            for d in dirnames:
                if d == prefix + str(image[id_field]):
                    image_path = path.join(dirpath, d)
                    found = True
                    break
            if found:
                break

        dicom = True
        for (dirpath, dirnames, filenames) in walk(image_path):
            for f in filenames:
                if f.endswith(".nii"):
                    dicom = False
                    image_path = path.join(dirpath, f)
                    break

        is_dicom.append(dicom)
        image_folders.append(image_path)
        if image_path == '':
            cprint('No %s image path found for subject %s in visit %s with image id %s'
                   % (modality, str(image.Subject_ID), str(image.VISCODE), str(image.Image_ID)))

    images.loc[:, 'Is_Dicom'] = pd.Series(is_dicom, index=images.index)
    images.loc[:, 'Path'] = pd.Series(image_folders, index=images.index)

    return images


def paths_to_bids(images, bids_dir, modality, mod_to_update=False):
    """
    Images in the list are converted and copied to directory in BIDS format

    Args:
        images: List of images metadata and paths
        bids_dir: Path to the output BIDS directory
        modality: Imaging modality
        mod_to_update:

    """
    from multiprocessing import Pool, cpu_count, Value
    from functools import partial

    if modality.lower() not in ['t1', 'dwi', 'flair', 'fmri', 'fdg', 'pib', 'av45_fbb', 'tau']:
        # This should never be reached
        raise RuntimeError(modality.lower()
                           + ' is not supported for conversion in paths_to_bids')

    counter = None

    def init(args):
        """
        Counter is stored for later use
        Args:
            args:
        """

        global counter
        counter = args

    counter = Value('i', 0)
    total = images.shape[0]
    # Reshape inputs to give it as a list to the workers
    images_list = []
    for i in range(total):
        images_list.append(images.iloc[i])

    # Handle multiargument for create_file functions
    partial_create_file = partial(create_file,
                                  modality=modality,
                                  total=total,
                                  bids_dir=bids_dir,
                                  mod_to_update=mod_to_update)

    # intializer are used with the counter variable to keep track of how many
    # files have been processed
    poolrunner = Pool(max(cpu_count() - 1, 1), initializer=init, initargs=(counter,))
    output_file_treated = poolrunner.map(partial_create_file, images_list)
    del counter
    return output_file_treated


def create_file(image, modality, total, bids_dir, mod_to_update):
    """
    Image file is created at the corresponding output folder
    as result of image conversion (DICOM to NIFTI) and centering,
    as specified for each modality

    Args:
        image: Image metadata
        modality: Imaging modality
        total: Total number of images to convert
        bids_dir: Path to the output BIDS directory
        mod_to_update:

    Returns:

    """
    import subprocess
    from colorama import Fore
    from clinica.utils.stream import cprint
    from clinica.iotools.utils.data_handling import center_nifti_origin
    from numpy import nan
    import os
    from os import path
    from glob import glob
    import shutil
    import re

    modality_specific = {'t1': {'output_path': 'anat',
                                'output_filename': '_T1w',
                                'to_center': True,
                                'json': 'n'},
                         'dwi': {'output_path': 'dwi',
                                 'output_filename': '_acq-axial_dwi',
                                 'to_center': False,
                                 'json': 'y'},
                         'flair': {'output_path': 'anat',
                                   'output_filename': '_FLAIR',
                                   'to_center': True,
                                   'json': 'y'},
                         'fmri': {'output_path': 'func',
                                  'output_filename': '_task-rest_bold',
                                  'to_center': False,
                                  'json': 'y'},
                         'fdg': {'output_path': 'pet',
                                 'output_filename': '_task-rest_acq-fdg_pet',
                                 'to_center': True,
                                 'json': 'n'},
                         'pib': {'output_path': 'pet',
                                 'output_filename': '_task-rest_acq-pib_pet',
                                 'to_center': True,
                                 'json': 'n'},
                         'av45': {'output_path': 'pet',
                                  'output_filename': '_task-rest_acq-av45_pet',
                                  'to_center': True,
                                  'json': 'n'},
                         'fbb': {'output_path': 'pet',
                                 'output_filename': '_task-rest_acq-fbb_pet',
                                 'to_center': True,
                                 'json': 'n'},
                         'tau': {'output_path': 'pet',
                                 'output_filename': '_task-rest_acq-tau_pet',
                                 'to_center': True,
                                 'json': 'n'}
                         }

    global counter

    with counter.get_lock():
        counter.value += 1

    subject = image.Subject_ID

    if modality == 'av45_fbb':
        modality = image.Tracer.lower()

    if image.Path == '':
        cprint(Fore.RED + '[' + modality.upper() + '] No path specified for '
               + image.Subject_ID + ' in session '
               + image.VISCODE + ' ' + str(counter.value)
               + ' / ' + str(total) + Fore.RESET)
        return nan
    cprint('[' + modality.upper() + '] Processing subject ' + str(subject)
           + ' - session ' + image.VISCODE + ', ' + str(counter.value)
           + ' / ' + str(total))
    session = viscode_to_session(image.VISCODE)
    image_path = image.Path
    image_id = image.Image_ID
    # If the original image is a DICOM, check if contains two DICOM
    # inside the same folder
    if image.Is_Dicom:
        image_path = check_two_dcm_folder(image_path, bids_dir, image_id)
    bids_subj = subject.replace('_', '')
    output_path = path.join(bids_dir, 'sub-ADNI' + bids_subj, 'ses-'
                            + session, modality_specific[modality]['output_path'])
    output_filename = 'sub-ADNI' + bids_subj + '_ses-' + session + modality_specific[modality]['output_filename']

    # If updated mode is selected, check if an old image is existing and remove it
    existing_im = glob(path.join(output_path, output_filename + '*'))
    if mod_to_update and len(existing_im) > 0:
        print('Removing the old image...')
        for im in existing_im:
            os.remove(im)

    try:
        os.makedirs(output_path)
    except OSError:
        # Folder already created with previous instance
        pass

    generate_json = modality_specific[modality]['json']
    if modality_specific[modality]['to_center']:
        zip_image = 'n'
    else:
        zip_image = 'y'

    if image.Is_Dicom:
        command = 'dcm2niix -b %s -z %s -o %s -f %s %s' % \
                  (generate_json, zip_image, output_path, output_filename, image_path)
        subprocess.run(command,
                       shell=True,
                       stderr=subprocess.DEVNULL,
                       stdout=subprocess.DEVNULL)

        # If "_t" - the trigger delay time - exists in dcm2niix output filename, we remove it
        exception_t = glob(path.join(output_path, output_filename + '_t[0-9]*'))
        for trigger_time in exception_t:
            res = re.search('_t\d+\.', trigger_time)
            no_trigger_time = trigger_time.replace(trigger_time[res.start(): res.end()], '.')
            os.rename(trigger_time, no_trigger_time)

        # Removing ADC images if one is generated
        adc_image = path.join(output_path, output_filename + '_ADC.nii.gz')
        if os.path.exists(adc_image):
            os.remove(adc_image)

        nifti_file = path.join(output_path, output_filename + '.nii')
        output_image = nifti_file + '.gz'

        # Conditions to check if output NIFTI files exists,
        # and, if DWI, if .bvec and .bval files are also present
        nifti_exists = path.isfile(nifti_file) or path.isfile(output_image)
        dwi_bvec_and_bval_exist = not (modality == 'dwi') or (path.isfile(path.join(output_path, output_filename + '.bvec'))
                                                              and path.isfile(path.join(output_path, output_filename + '.bval')))

        # Check if conversion worked (output files exist)
        if not nifti_exists or not dwi_bvec_and_bval_exist:

            cprint('WARNING: Conversion with dcm2niix failed, trying with dcm2nii '
                   'for subject ' + subject + ' and session ' + session)
            command = 'dcm2nii -a n -d n -e n -i y -g %s -p n -m n -r n -x n -o %s %s' % \
                      (zip_image, output_path, image_path)
            subprocess.run(command,
                           shell=True,
                           stdout=subprocess.DEVNULL,
                           stderr=subprocess.DEVNULL)

            # If modality is DWI we check if .bvec and .bval files are also present
            if modality == "dwi":
                bvec_file = path.join(output_path, subject.replace('_', '') + '.bvec')
                bval_file = path.join(output_path, subject.replace('_', '') + '.bval')

                if os.path.exists(bvec_file) and os.path.exists(bval_file):
                    os.rename(bvec_file, path.join(output_path, output_filename + '.bvec'))
                    os.rename(bval_file, path.join(output_path, output_filename + '.bval'))
                else:
                    cprint(Fore.RED + 'WARNING: bvec and bval not generated by dcm2nii '
                           'for subject ' + subject + ' and session ' + session + '. ' + Fore.RESET)

            nifti_files = glob(path.join(output_path, subject.replace('_', '') + '.nii*'))
            output_image = path.join(output_path, output_filename + '.nii.gz')

            # If no NIFTI files were converted then conversion failed
            if nifti_files:
                nifti_file = nifti_files[0]
                conversion_failed = False
            else:
                conversion_failed = True

            # If image is not going to be centered, then it is already a .nii.gz file
            # and we only need to rename it to the output image filename
            if not conversion_failed and not modality_specific[modality]['to_center']:
                os.rename(nifti_file, output_image)

            # We check if neither the NIFTI file to be centered
            # nor the final compressed NIFTI file exist. In this case, conversion failed
            if conversion_failed or (not path.isfile(nifti_file) and not path.isfile(output_image)):
                cprint(Fore.RED + 'WARNING: Conversion of the dicom failed '
                       'for subject ' + subject + ' and session ' + session + '. '
                       'Image path: ' + image_path + Fore.RESET)
                # If conversion failed we remove the folder if it is empty
                if not os.listdir(output_path):
                    os.rmdir(output_path)
                return nan

        # Case when JSON file was expected, but not generated by dcm2niix
        elif generate_json == 'y' and not os.path.exists(path.join(output_path, output_filename + '.json')):
            cprint('WARNING: JSON file not generated by dcm2niix '
                   'for subject ' + subject + ' and session ' + session)

        if modality_specific[modality]['to_center']:
            center_nifti_origin(nifti_file, output_image)
            os.remove(nifti_file)

    else:
        output_image = path.join(output_path, output_filename + '.nii.gz')
        if modality_specific[modality]['to_center']:
            center_nifti_origin(image_path, output_image)
            if output_image is None:
                cprint("Error: For subject %s in session%s an error occurred recentering Nifti image:%s"
                       % (subject, session, image_path))
        else:
            shutil.copy(image_path, output_image)

    # Check if there is still the folder tmp_dcm_folder and remove it
    remove_tmp_dmc_folder(bids_dir, image_id)
    return output_image


def viscode_to_session(viscode):
    """
    Replace the session label 'bl' with 'M00' or capitalize the session name passed
    as input

    Args:
        viscode: session name

    Returns: M00 if is the baseline session or the original session name capitalized

    """

    if viscode == 'bl':
        return 'M00'
    else:
        return viscode.capitalize()


def check_two_dcm_folder(dicom_path, bids_folder, image_uid):
    """
    Check if a folder contains more than one DICOM and if yes, copy the DICOM related to image id passed as parameter
    into a temporary folder called tmp_dicom_folder

    Args:
        dicom_path: path to the DICOM folder
        bids_folder: path to the BIDS folder where the dataset will be stored
        image_uid: image id of the fmri

    Returns: the path to the original DICOM folder or the path to a temporary folder called tmp_dicom_folder where only
     the DICOM to convert is copied

    """

    from glob import glob
    from os import path
    from shutil import copy
    import shutil
    import os

    temp_folder_name = 'tmp_dcm_folder_' + str(image_uid).strip(' ')
    dest_path = path.join(bids_folder, temp_folder_name)

    # Check if there are dicom files inside the folder not belonging to the desired image
    dicom_list = glob(path.join(dicom_path, '*.dcm'))
    image_list = glob(path.join(dicom_path, '*%s.dcm' % image_uid))
    if len(dicom_list) != len(image_list):

        # Remove the precedent tmp_dcm_folder if is existing
        if os.path.exists(dest_path):
            shutil.rmtree(dest_path)
        os.mkdir(dest_path)
        dmc_to_conv = glob(path.join(dicom_path, '*'+str(image_uid)+'.dcm*'))
        for d in dmc_to_conv:
            copy(d, dest_path)
        return dest_path
    else:
        return dicom_path


def remove_tmp_dmc_folder(bids_dir, image_id):
    """
    Remove the folder tmp_dmc_folder created by the method check_two_dcm_folder (if existing)

    Args:
        bids_dir: path to the BIDS directory
        image_id:

    Returns:

    """
    from os.path import exists, join
    from shutil import rmtree

    tmp_dcm_folder_path = join(bids_dir, 'tmp_dcm_folder_' + str(image_id).strip(' '))
    if exists(tmp_dcm_folder_path):
        rmtree(tmp_dcm_folder_path)



# coding: utf8


from abc import ABC, abstractmethod

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class MLWorkflow(ABC):

    def __init__(self, input_class, validation_class, algorithm_class, all_params, output_dir):

        self._input_class = input_class
        self._validation_class = validation_class
        self._algorithm_class = algorithm_class

        self._input_params = self.create_parameters_dict(all_params, input_class)
        self._validation_params = self.create_parameters_dict(all_params, validation_class)
        self._algorithm_params = self.create_parameters_dict(all_params, algorithm_class)

        self._output_dir = output_dir

        self._input = None
        self._validation = None
        self._algorithm = None

    def run(self):

        from os import path, makedirs

        # Instantiating input class
        self._input = self._input_class(self._input_params)

        # Computing input values
        x = self._input.get_x()
        y = self._input.get_y()

        # Instantiating classification algorithm
        if self._algorithm_class.uses_kernel():
            kernel = self._input.get_kernel()
            self._algorithm = self._algorithm_class(kernel, y, self._algorithm_params)
        else:
            self._algorithm = self._algorithm_class(x, y, self._algorithm_params)

        # Instantiating cross-validation method and classification algorithm
        self._validation = self._validation_class(self._algorithm, self._validation_params)

        # Launching classification with selected cross-validation
        classifier, best_params, results = self._validation.validate(y)

        # Creation of the directory to save results
        classifier_dir = path.join(self._output_dir, 'classifier')
        if not path.exists(classifier_dir):
            makedirs(classifier_dir)

        # Saving algorithm trained classifier
        self._algorithm.save_classifier(classifier, classifier_dir)
        self._algorithm.save_weights(classifier, x, classifier_dir)
        self._algorithm.save_parameters(best_params, classifier_dir)

        # Saving validation trained classifier
        self._validation.save_results(self._output_dir)

    @staticmethod
    def create_parameters_dict(locals_dictionary, component_class):

        default_parameters = component_class.get_default_parameters()
        for key in locals_dictionary:
            if key in default_parameters:
                default_parameters[key] = locals_dictionary[key]
        return default_parameters


class MLInput(ABC):

    def __init__(self, input_params):

        self._input_params = self.get_default_parameters()
        self._input_params.update(input_params)

        self._x = None
        self._y = None
        self._kernel = None

    @abstractmethod
    def get_x(self):
        pass

    @abstractmethod
    def get_y(self):
        pass

    @staticmethod
    @abstractmethod
    def get_default_parameters():
        pass


class MLValidation(ABC):

    def __init__(self, ml_algorithm, validation_params):

        self._ml_algorithm = ml_algorithm

        self._validation_params = self.get_default_parameters()
        self._validation_params.update(validation_params)

        self._validation_results = []
        self._classifier = None
        self._best_params = None

    @abstractmethod
    def validate(self, y):
        pass

    @staticmethod
    @abstractmethod
    def get_default_parameters():
        pass


class MLAlgorithm(ABC):

    def __init__(self, input_data, y, algorithm_params):

        self._algorithm_params = self.get_default_parameters()
        self._algorithm_params.update(algorithm_params)

        if self.uses_kernel():
            self._kernel = input_data
        else:
            self._x = input_data

        self._y = y

    @staticmethod
    @abstractmethod
    def uses_kernel():
        pass

    @abstractmethod
    def evaluate(self, train_index, test_index):
        pass

    @abstractmethod
    def save_classifier(self, classifier, output_dir):
        pass

    @abstractmethod
    def save_parameters(self, parameters, output_dir):
        pass

    @staticmethod
    @abstractmethod
    def get_default_parameters():
        pass
